#!/usr/bin/env python
# Copyright 2021-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
TDA and TDHF for no-pair DKS Hamiltonian
'''

from functools import reduce
import numpy
from pyscf import lib
from pyscf import scf
from pyscf import ao2mo
from pyscf.lib import logger
from pyscf.tdscf import rhf
from pyscf.tdscf._lr_eig import eigh as lr_eigh, eig as lr_eig
from pyscf import __config__

OUTPUT_THRESHOLD = getattr(__config__, 'tdscf_uhf_get_nto_threshold', 0.3)
REAL_EIG_THRESHOLD = getattr(__config__, 'tdscf_uhf_TDDFT_pick_eig_threshold', 1e-4)

def gen_tda_operation(mf, fock_ao=None, with_nlc=True):
    '''A x
    '''
gen_tda_hop = gen_tda_operation

def _gen_tda_operation(td, fock_ao=None):
    assert fock_ao is None
    mf = td._scf
    mo_coeff = mf.mo_coeff
    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    nao, nmo = mo_coeff.shape
    # Remove all negative states
    n2c = nmo // 2
    occidx = n2c + numpy.where(mo_occ[n2c:] == 1)[0]
    viridx = n2c + numpy.where(mo_occ[n2c:] == 0)[0]
    nocc = len(occidx)
    nvir = len(viridx)
    orbv = mo_coeff[:,viridx]
    orbo = mo_coeff[:,occidx]

    e_ia = hdiag = mo_energy[viridx] - mo_energy[occidx,None]
    hdiag = hdiag.ravel()

    mo_coeff = numpy.asarray(numpy.hstack((orbo,orbv)), order='F')
    vresp = td.gen_response(hermi=0)

    def vind(zs):
        zs = numpy.asarray(zs).reshape(-1,nocc,nvir)
        dms = lib.einsum('xov,pv,qo->xpq', zs, orbv, orbo.conj())
        v1ao = vresp(dms)
        v1mo = lib.einsum('xpq,qo,pv->xov', v1ao, orbo, orbv.conj())
        v1mo += numpy.einsum('xia,ia->xia', zs, e_ia)
        return v1mo.reshape(v1mo.shape[0], -1)

    return vind, hdiag

def get_ab(mf, mo_energy=None, mo_coeff=None, mo_occ=None):
    r'''A and B matrices for TDDFT response function.

    A[i,a,j,b] = \delta_{ab}\delta_{ij}(E_a - E_i) + (ai||jb)
    B[i,a,j,b] = (ai||bj)
    '''
    if mo_energy is None: mo_energy = mf.mo_energy
    if mo_coeff is None: mo_coeff = mf.mo_coeff
    if mo_occ is None: mo_occ = mf.mo_occ

    mol = mf.mol
    nao, nmo = mo_coeff.shape
    n2c = nmo // 2
    occidx = n2c + numpy.where(mo_occ[n2c:] == 1)[0]
    viridx = n2c + numpy.where(mo_occ[n2c:] == 0)[0]
    orbv = mo_coeff[:,viridx]
    orbo = mo_coeff[:,occidx]
    nvir = orbv.shape[1]
    nocc = orbo.shape[1]
    nmo = nocc + nvir
    mo = numpy.hstack((orbo, orbv))
    c1 = .5 / lib.param.LIGHT_SPEED
    moL = numpy.asarray(mo[:n2c], order='F')
    moS = numpy.asarray(mo[n2c:], order='F') * c1
    orboL = moL[:,:nocc]
    orboS = moS[:,:nocc]

    e_ia = mo_energy[viridx] - mo_energy[occidx,None]
    a = numpy.diag(e_ia.ravel()).reshape(nocc,nvir,nocc,nvir)
    b = numpy.zeros_like(a)

    def add_hf_(a, b, hyb=1):
        eri_mo = ao2mo.kernel(mol, [orboL, moL, moL, moL], intor='int2e_spinor')
        eri_mo+= ao2mo.kernel(mol, [orboS, moS, moS, moS], intor='int2e_spsp1spsp2_spinor')
        eri_mo+= ao2mo.kernel(mol, [orboS, moS, moL, moL], intor='int2e_spsp1_spinor')
        eri_mo+= ao2mo.kernel(mol, [moS, moS, orboL, moL], intor='int2e_spsp1_spinor').T
        eri_mo = eri_mo.reshape(nocc,nmo,nmo,nmo)

        a = a + numpy.einsum('iabj->iajb', eri_mo[:nocc,nocc:,nocc:,:nocc].conj())
        a = a - numpy.einsum('ijba->iajb', eri_mo[:nocc,:nocc,nocc:,nocc:].conj()) * hyb
        b = b + numpy.einsum('iajb->iajb', eri_mo[:nocc,nocc:,:nocc,nocc:].conj())
        b = b - numpy.einsum('jaib->iajb', eri_mo[:nocc,nocc:,:nocc,nocc:].conj()) * hyb
        return a, b

    if isinstance(mf, scf.hf.KohnShamDFT):
        from pyscf.dft import xc_deriv
        ni = mf._numint
        ni.libxc.test_deriv_order(mf.xc, 2, raise_error=True)
        if mf.do_nlc():
            raise NotImplementedError('DKS-TDDFT for NLC functionals')

        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, mol.spin)
        assert omega == 0.

        a, b = add_hf_(a, b, hyb)

        xctype = ni._xc_type(mf.xc)
        dm0 = mf.make_rdm1(mo_coeff, mo_occ)
        mem_now = lib.current_memory()[0]
        max_memory = max(2000, mf.max_memory*.8-mem_now)

        def get_mo_value(ao):
            aoLa, aoLb, aoSa, aoSb = ao
            if aoLa.ndim == 2:
                moLa = lib.einsum('rp,pi->ri', aoLa, moL)
                moLb = lib.einsum('rp,pi->ri', aoLb, moL)
                moSa = lib.einsum('rp,pi->ri', aoSa, moS)
                moSb = lib.einsum('rp,pi->ri', aoSb, moS)
                return (moLa[:,:nocc], moLa[:,nocc:], moLb[:,:nocc], moLb[:,nocc:],
                        moSa[:,:nocc], moSa[:,nocc:], moSb[:,:nocc], moSb[:,nocc:])
            else:
                moLa = lib.einsum('xrp,pi->xri', aoLa, moL)
                moLb = lib.einsum('xrp,pi->xri', aoLb, moL)
                moSa = lib.einsum('xrp,pi->xri', aoSa, moS)
                moSb = lib.einsum('xrp,pi->xri', aoSb, moS)
                return (moLa[:,:,:nocc], moLa[:,:,nocc:], moLb[:,:,:nocc], moLb[:,:,nocc:],
                        moSa[:,:,:nocc], moSa[:,:,nocc:], moSb[:,:,:nocc], moSb[:,:,nocc:])

        def ud2tm(aa, ab, ba, bb):
            return numpy.stack([aa + bb,        # rho
                                ba + ab,        # mx
                                (ba - ab) * 1j, # my
                                aa - bb])       # mz
        def addLS(rhoL, rhoS):
            rhoS[1:4] *= -1  # beta * Sigma
            return rhoL + rhoS

        if xctype == 'LDA':
            ao_deriv = 0
            for ao, mask, weight, coords \
                    in ni.block_loop(mol, mf.grids, nao, ao_deriv, max_memory,
                                     with_s=True):
                if ni.collinear[0] == 'm':
                    rho = ni.eval_rho(mol, ao, dm0, mask, xctype, hermi=1, with_lapl=False)
                    eval_xc = ni.mcfun_eval_xc_adapter(mf.xc)
                    fxc = eval_xc(mf.xc, rho, deriv=2, xctype=xctype)[2]
                    wfxc = weight * fxc.reshape(4,4,-1)
                    moLoa, moLva, moLob, moLvb, moSoa, moSva, moSob, moSvb = get_mo_value(ao)
                    rhoL_ov_aa = numpy.einsum('ri,ra->ria', moLoa.conj(), moLva)
                    rhoL_ov_ab = numpy.einsum('ri,ra->ria', moLoa.conj(), moLvb)
                    rhoL_ov_ba = numpy.einsum('ri,ra->ria', moLob.conj(), moLva)
                    rhoL_ov_bb = numpy.einsum('ri,ra->ria', moLob.conj(), moLvb)
                    rhoS_ov_aa = numpy.einsum('ri,ra->ria', moSoa.conj(), moSva)
                    rhoS_ov_ab = numpy.einsum('ri,ra->ria', moSoa.conj(), moSvb)
                    rhoS_ov_ba = numpy.einsum('ri,ra->ria', moSob.conj(), moSva)
                    rhoS_ov_bb = numpy.einsum('ri,ra->ria', moSob.conj(), moSvb)
                    rhoL_ov = ud2tm(rhoL_ov_aa, rhoL_ov_ab, rhoL_ov_ba, rhoL_ov_bb)
                    rhoS_ov = ud2tm(rhoS_ov_aa, rhoS_ov_ab, rhoS_ov_ba, rhoS_ov_bb)
                    rho_ov = addLS(rhoL_ov, rhoS_ov)
                    rho_vo = rho_ov.conj()
                    w_vo = numpy.einsum('tsr,tria->sria', wfxc, rho_vo)
                    a += lib.einsum('sria,srjb->iajb', w_vo, rho_ov)
                    b += lib.einsum('sria,srjb->iajb', w_vo, rho_vo)
                elif ni.collinear[0] == 'c':
                    rho = ni.eval_rho(mol, ao, dm0, mask, xctype, hermi=1, with_lapl=False)
                    fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2)[2]
                    wv_a, wv_b = weight * fxc.reshape(2,2,-1)
                    moLoa, moLva, moLob, moLvb, moSoa, moSva, moSob, moSvb = get_mo_value(ao)
                    rhoL_ov_a = numpy.einsum('ri,ra->ria', moLoa.conj(), moLva)
                    rhoL_ov_b = numpy.einsum('ri,ra->ria', moLob.conj(), moLvb)
                    rhoS_ov_a = numpy.einsum('ri,ra->ria', moSoa.conj(), moSva)
                    rhoS_ov_b = numpy.einsum('ri,ra->ria', moSob.conj(), moSvb)
                    rhoL_vo_a = rhoL_ov_a.conj()
                    rhoL_vo_b = rhoL_ov_b.conj()
                    rhoS_vo_a = rhoS_ov_a.conj()
                    rhoS_vo_b = rhoS_ov_b.conj()
                    w_vo  = wv_a[:,:,None,None] * rhoL_vo_a
                    w_vo += wv_b[:,:,None,None] * rhoL_vo_b
                    w_vo += wv_b[:,:,None,None] * rhoS_vo_a  # for beta*Sigma
                    w_vo += wv_a[:,:,None,None] * rhoS_vo_b
                    wa_vo, wb_vo = w_vo
                    a += lib.einsum('ria,rjb->iajb', wa_vo, rhoL_ov_a)
                    a += lib.einsum('ria,rjb->iajb', wb_vo, rhoL_ov_b)
                    a += lib.einsum('ria,rjb->iajb', wb_vo, rhoS_ov_a)
                    a += lib.einsum('ria,rjb->iajb', wa_vo, rhoS_ov_b)
                    b += lib.einsum('ria,rjb->iajb', wa_vo, rhoL_vo_a)
                    b += lib.einsum('ria,rjb->iajb', wb_vo, rhoL_vo_b)
                    b += lib.einsum('ria,rjb->iajb', wb_vo, rhoS_vo_a)
                    b += lib.einsum('ria,rjb->iajb', wa_vo, rhoS_vo_b)
                else:
                    raise NotImplementedError(ni.collinear)

        elif xctype == 'GGA':
            ao_deriv = 1
            for ao, mask, weight, coords \
                    in ni.block_loop(mol, mf.grids, nao, ao_deriv, max_memory,
                                     with_s=True):
                if ni.collinear[0] == 'm':
                    rho = ni.eval_rho(mol, ao, dm0, mask, xctype, hermi=1, with_lapl=False)
                    eval_xc = ni.mcfun_eval_xc_adapter(mf.xc)
                    fxc = eval_xc(mf.xc, rho, deriv=2, xctype=xctype)[2]
                    wfxc = weight * fxc
                    moLoa, moLva, moLob, moLvb, moSoa, moSva, moSob, moSvb = get_mo_value(ao)
                    rhoL_ov_aa = numpy.einsum('ri,xra->xria', moLoa[0].conj(), moLva)
                    rhoL_ov_ab = numpy.einsum('ri,xra->xria', moLoa[0].conj(), moLvb)
                    rhoL_ov_ba = numpy.einsum('ri,xra->xria', moLob[0].conj(), moLva)
                    rhoL_ov_bb = numpy.einsum('ri,xra->xria', moLob[0].conj(), moLvb)
                    rhoS_ov_aa = numpy.einsum('ri,xra->xria', moSoa[0].conj(), moSva)
                    rhoS_ov_ab = numpy.einsum('ri,xra->xria', moSoa[0].conj(), moSvb)
                    rhoS_ov_ba = numpy.einsum('ri,xra->xria', moSob[0].conj(), moSva)
                    rhoS_ov_bb = numpy.einsum('ri,xra->xria', moSob[0].conj(), moSvb)
                    rhoL_ov_aa[1:4] += numpy.einsum('xri,ra->xria', moLoa[1:4].conj(), moLva[0])
                    rhoL_ov_ab[1:4] += numpy.einsum('xri,ra->xria', moLoa[1:4].conj(), moLvb[0])
                    rhoL_ov_ba[1:4] += numpy.einsum('xri,ra->xria', moLob[1:4].conj(), moLva[0])
                    rhoL_ov_bb[1:4] += numpy.einsum('xri,ra->xria', moLob[1:4].conj(), moLvb[0])
                    rhoS_ov_aa[1:4] += numpy.einsum('xri,ra->xria', moSoa[1:4].conj(), moSva[0])
                    rhoS_ov_ab[1:4] += numpy.einsum('xri,ra->xria', moSoa[1:4].conj(), moSvb[0])
                    rhoS_ov_ba[1:4] += numpy.einsum('xri,ra->xria', moSob[1:4].conj(), moSva[0])
                    rhoS_ov_bb[1:4] += numpy.einsum('xri,ra->xria', moSob[1:4].conj(), moSvb[0])
                    rhoL_ov = ud2tm(rhoL_ov_aa, rhoL_ov_ab, rhoL_ov_ba, rhoL_ov_bb)
                    rhoS_ov = ud2tm(rhoS_ov_aa, rhoS_ov_ab, rhoS_ov_ba, rhoS_ov_bb)
                    rho_ov = addLS(rhoL_ov, rhoS_ov)
                    rho_vo = rho_ov.conj()
                    w_vo = numpy.einsum('txsyr,txria->syria', wfxc, rho_vo)
                    a += lib.einsum('syria,syrjb->iajb', w_vo, rho_ov)
                    b += lib.einsum('syria,syrjb->iajb', w_vo, rho_vo)
                elif ni.collinear[0] == 'c':
                    rho = ni.eval_rho(mol, ao, dm0, mask, xctype, hermi=1, with_lapl=False)
                    fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2)[2]
                    fxc = xc_deriv.ud2ts(fxc)
                    wfxc = weight * fxc
                    moLoa, moLva, moLob, moLvb, moSoa, moSva, moSob, moSvb = get_mo_value(ao)
                    rhoL_ov_a = numpy.einsum('xri,ra->xria', moLoa.conj(), moLva[0])
                    rhoL_ov_b = numpy.einsum('xri,ra->xria', moLob.conj(), moLvb[0])
                    rhoL_ov_a[1:4] += numpy.einsum('ri,xra->xria', moLoa[0].conj(), moLva[1:4])
                    rhoL_ov_b[1:4] += numpy.einsum('ri,xra->xria', moLob[0].conj(), moLvb[1:4])
                    rhoS_ov_a = numpy.einsum('xri,ra->xria', moSoa.conj(), moSva[0])
                    rhoS_ov_b = numpy.einsum('xri,ra->xria', moSob.conj(), moSvb[0])
                    rhoS_ov_a[1:4] += numpy.einsum('ri,xra->xria', moSoa[0].conj(), moSva[1:4])
                    rhoS_ov_b[1:4] += numpy.einsum('ri,xra->xria', moSob[0].conj(), moSvb[1:4])
                    rhoL_ov = numpy.stack((rhoL_ov_a+rhoL_ov_b, rhoL_ov_a-rhoL_ov_b))
                    rhoS_ov = numpy.stack((rhoS_ov_a+rhoS_ov_b, rhoS_ov_a-rhoS_ov_b))
                    rhoS_ov[1] *= -1
                    rho_ov = rhoL_ov + rhoS_ov
                    rho_vo = rho_ov.conj()
                    w_vo = numpy.einsum('txsyr,txria->syria', wfxc, rho_vo)
                    a += lib.einsum('syria,syrjb->iajb', w_vo, rho_ov)
                    b += lib.einsum('syria,syrjb->iajb', w_vo, rho_vo)
                else:
                    raise NotImplementedError(ni.collinear)

        elif xctype == 'HF':
            pass

        elif xctype == 'NLC':
            raise NotImplementedError('NLC')

        elif xctype == 'MGGA':
            ao_deriv = 1
            for ao, mask, weight, coords \
                    in ni.block_loop(mol, mf.grids, nao, ao_deriv, max_memory,
                                     with_s=True):
                if ni.collinear[0] == 'm':
                    rho = ni.eval_rho(mol, ao, dm0, mask, xctype, hermi=1, with_lapl=False)
                    eval_xc = ni.mcfun_eval_xc_adapter(mf.xc)
                    fxc = eval_xc(mf.xc, rho, deriv=2, xctype=xctype)[2]
                    wfxc = weight * fxc
                    moLoa, moLva, moLob, moLvb, moSoa, moSva, moSob, moSvb = get_mo_value(ao)
                    rhoL_ov_aa = numpy.einsum('ri,xra->xria', moLoa[0].conj(), moLva)
                    rhoL_ov_ab = numpy.einsum('ri,xra->xria', moLoa[0].conj(), moLvb)
                    rhoL_ov_ba = numpy.einsum('ri,xra->xria', moLob[0].conj(), moLva)
                    rhoL_ov_bb = numpy.einsum('ri,xra->xria', moLob[0].conj(), moLvb)
                    rhoS_ov_aa = numpy.einsum('ri,xra->xria', moSoa[0].conj(), moSva)
                    rhoS_ov_ab = numpy.einsum('ri,xra->xria', moSoa[0].conj(), moSvb)
                    rhoS_ov_ba = numpy.einsum('ri,xra->xria', moSob[0].conj(), moSva)
                    rhoS_ov_bb = numpy.einsum('ri,xra->xria', moSob[0].conj(), moSvb)
                    rhoL_ov_aa[1:4] += numpy.einsum('xri,ra->xria', moLoa[1:4].conj(), moLva[0])
                    rhoL_ov_ab[1:4] += numpy.einsum('xri,ra->xria', moLoa[1:4].conj(), moLvb[0])
                    rhoL_ov_ba[1:4] += numpy.einsum('xri,ra->xria', moLob[1:4].conj(), moLva[0])
                    rhoL_ov_bb[1:4] += numpy.einsum('xri,ra->xria', moLob[1:4].conj(), moLvb[0])
                    rhoS_ov_aa[1:4] += numpy.einsum('xri,ra->xria', moSoa[1:4].conj(), moSva[0])
                    rhoS_ov_ab[1:4] += numpy.einsum('xri,ra->xria', moSoa[1:4].conj(), moSvb[0])
                    rhoS_ov_ba[1:4] += numpy.einsum('xri,ra->xria', moSob[1:4].conj(), moSva[0])
                    rhoS_ov_bb[1:4] += numpy.einsum('xri,ra->xria', moSob[1:4].conj(), moSvb[0])
                    tauL_ov_aa = numpy.einsum('xri,xra->ria', moLoa[1:4].conj(), moLva[1:4]) * .5
                    tauL_ov_ab = numpy.einsum('xri,xra->ria', moLoa[1:4].conj(), moLvb[1:4]) * .5
                    tauL_ov_ba = numpy.einsum('xri,xra->ria', moLob[1:4].conj(), moLva[1:4]) * .5
                    tauL_ov_bb = numpy.einsum('xri,xra->ria', moLob[1:4].conj(), moLvb[1:4]) * .5
                    tauS_ov_aa = numpy.einsum('xri,xra->ria', moSoa[1:4].conj(), moSva[1:4]) * .5
                    tauS_ov_ab = numpy.einsum('xri,xra->ria', moSoa[1:4].conj(), moSvb[1:4]) * .5
                    tauS_ov_ba = numpy.einsum('xri,xra->ria', moSob[1:4].conj(), moSva[1:4]) * .5
                    tauS_ov_bb = numpy.einsum('xri,xra->ria', moSob[1:4].conj(), moSvb[1:4]) * .5
                    rhoL_ov_aa = numpy.vstack([rhoL_ov_aa, tauL_ov_aa[numpy.newaxis]])
                    rhoL_ov_ab = numpy.vstack([rhoL_ov_ab, tauL_ov_ab[numpy.newaxis]])
                    rhoL_ov_ba = numpy.vstack([rhoL_ov_ba, tauL_ov_ba[numpy.newaxis]])
                    rhoL_ov_bb = numpy.vstack([rhoL_ov_bb, tauL_ov_bb[numpy.newaxis]])
                    rhoS_ov_aa = numpy.vstack([rhoS_ov_aa, tauS_ov_aa[numpy.newaxis]])
                    rhoS_ov_ab = numpy.vstack([rhoS_ov_ab, tauS_ov_ab[numpy.newaxis]])
                    rhoS_ov_ba = numpy.vstack([rhoS_ov_ba, tauS_ov_ba[numpy.newaxis]])
                    rhoS_ov_bb = numpy.vstack([rhoS_ov_bb, tauS_ov_bb[numpy.newaxis]])
                    rhoL_ov = ud2tm(rhoL_ov_aa, rhoL_ov_ab, rhoL_ov_ba, rhoL_ov_bb)
                    rhoS_ov = ud2tm(rhoS_ov_aa, rhoS_ov_ab, rhoS_ov_ba, rhoS_ov_bb)
                    rho_ov = addLS(rhoL_ov, rhoS_ov)
                    rho_vo = rho_ov.conj()
                    w_vo = numpy.einsum('txsyr,txria->syria', wfxc, rho_vo)
                    a += lib.einsum('syria,syrjb->iajb', w_vo, rho_ov)
                    b += lib.einsum('syria,syrjb->iajb', w_vo, rho_vo)
                elif ni.collinear[0] == 'c':
                    rho = ni.eval_rho(mol, ao, dm0, mask, xctype, hermi=1, with_lapl=False)
                    fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2)[2]
                    fxc = xc_deriv.ud2ts(fxc)
                    wfxc = weight * fxc
                    moLoa, moLva, moLob, moLvb, moSoa, moSva, moSob, moSvb = get_mo_value(ao)
                    rhoL_ov_a = numpy.einsum('xri,ra->xria', moLoa.conj(), moLva[0])
                    rhoL_ov_b = numpy.einsum('xri,ra->xria', moLob.conj(), moLvb[0])
                    rhoS_ov_a = numpy.einsum('xri,ra->xria', moSoa.conj(), moSva[0])
                    rhoS_ov_b = numpy.einsum('xri,ra->xria', moSob.conj(), moSvb[0])
                    rhoL_ov_a[1:4] += numpy.einsum('ri,xra->xria', moLoa[0].conj(), moLva[1:4])
                    rhoL_ov_b[1:4] += numpy.einsum('ri,xra->xria', moLob[0].conj(), moLvb[1:4])
                    rhoS_ov_a[1:4] += numpy.einsum('ri,xra->xria', moSoa[0].conj(), moSva[1:4])
                    rhoS_ov_b[1:4] += numpy.einsum('ri,xra->xria', moSob[0].conj(), moSvb[1:4])
                    tauL_ov_a = numpy.einsum('xri,xra->ria', moLoa[1:4].conj(), moLva[1:4]) * .5
                    tauL_ov_b = numpy.einsum('xri,xra->ria', moLob[1:4].conj(), moLvb[1:4]) * .5
                    tauS_ov_a = numpy.einsum('xri,xra->ria', moSoa[1:4].conj(), moSva[1:4]) * .5
                    tauS_ov_b = numpy.einsum('xri,xra->ria', moSob[1:4].conj(), moSvb[1:4]) * .5
                    rhoL_ov_a = numpy.vstack([rhoL_ov_a, tauL_ov_a[numpy.newaxis]])
                    rhoL_ov_b = numpy.vstack([rhoL_ov_b, tauL_ov_b[numpy.newaxis]])
                    rhoS_ov_a = numpy.vstack([rhoS_ov_a, tauS_ov_a[numpy.newaxis]])
                    rhoS_ov_b = numpy.vstack([rhoS_ov_b, tauS_ov_b[numpy.newaxis]])
                    rhoL_ov = numpy.stack((rhoL_ov_a+rhoL_ov_b, rhoL_ov_a-rhoL_ov_b))
                    rhoS_ov = numpy.stack((rhoS_ov_a+rhoS_ov_b, rhoS_ov_a-rhoS_ov_b))
                    rhoS_ov[1] *= -1
                    rho_ov = rhoL_ov + rhoS_ov
                    rho_vo = rho_ov.conj()
                    w_vo = numpy.einsum('txsyr,txria->syria', wfxc, rho_vo)
                    a += lib.einsum('syria,syrjb->iajb', w_vo, rho_ov)
                    b += lib.einsum('syria,syrjb->iajb', w_vo, rho_vo)
                else:
                    raise NotImplementedError(ni.collinear)

    else:
        a, b = add_hf_(a, b)

    return a, b

def get_nto(tdobj, state=1, threshold=OUTPUT_THRESHOLD, verbose=None):
    raise NotImplementedError('get_nto')

def analyze(tdobj, verbose=None):
    raise NotImplementedError('analyze')

def _contract_multipole(tdobj, ints, hermi=True, xy=None):
    raise NotImplementedError


class TDBase(rhf.TDBase):

    @lib.with_doc(get_ab.__doc__)
    def get_ab(self, mf=None):
        if mf is None: mf = self._scf
        return get_ab(mf)

    analyze = analyze
    get_nto = get_nto
    _contract_multipole = _contract_multipole  # needed by transition dipoles

    def nuc_grad_method(self):
        raise NotImplementedError


@lib.with_doc(rhf.TDA.__doc__)
class TDA(TDBase):

    singlet = None

    def gen_vind(self, mf=None):
        '''Generate function to compute Ax'''
        assert mf is None or mf is self._scf
        return _gen_tda_operation(self)

    def get_init_guess(self, mf, nstates=None, wfnsym=None):
        if nstates is None: nstates = self.nstates
        mo_energy = mf.mo_energy
        mo_occ = mf.mo_occ
        # Remove all negative states
        n2c = mf.mo_occ.size // 2
        occidx = n2c + numpy.where(mo_occ[n2c:] == 1)[0]
        viridx = n2c + numpy.where(mo_occ[n2c:] == 0)[0]
        e_ia = mo_energy[viridx] - mo_energy[occidx,None]

        nov = e_ia.size
        nstates = min(nstates, nov)
        e_ia = e_ia.ravel()
        e_threshold = numpy.sort(e_ia)[nstates-1]
        e_threshold += self.deg_eia_thresh

        idx = numpy.where(e_ia <= e_threshold)[0]
        x0 = numpy.zeros((idx.size, nov))
        for i, j in enumerate(idx):
            x0[i, j] = 1  # Koopmans' excitations
        return x0

    def kernel(self, x0=None, nstates=None):
        '''TDA diagonalization solver
        '''
        cpu0 = (logger.process_clock(), logger.perf_counter())
        self.check_sanity()
        self.dump_flags()
        if nstates is None:
            nstates = self.nstates
        else:
            self.nstates = nstates

        log = logger.Logger(self.stdout, self.verbose)

        vind, hdiag = self.gen_vind(self._scf)
        precond = self.get_precond(hdiag)

        def pickeig(w, v, nroots, envs):
            idx = numpy.where(w > self.positive_eig_threshold)[0]
            return w[idx], v[:,idx], idx

        if x0 is None:
            x0 = self.get_init_guess(self._scf, self.nstates)

        self.converged, self.e, x1 = lr_eigh(
            vind, x0, precond, tol_residual=self.conv_tol, lindep=self.lindep,
            nroots=nstates, pick=pickeig, max_cycle=self.max_cycle,
            max_memory=self.max_memory, verbose=log)

        nocc = (self._scf.mo_occ>0).sum()
        nmo = self._scf.mo_occ.size
        n2c = nmo // 2
        nvir = n2c - nocc
        self.xy = [(xi.reshape(nocc,nvir), 0) for xi in x1]

        if self.chkfile:
            lib.chkfile.save(self.chkfile, 'tddft/e', self.e)
            lib.chkfile.save(self.chkfile, 'tddft/xy', self.xy)

        log.timer('TDA', *cpu0)
        self._finalize()
        return self.e, self.xy


def gen_tdhf_operation(mf, fock_ao=None, with_nlc=True):
    '''Generate function to compute

    [ A   B ][X]
    [-B* -A*][Y]
    '''
    td = TDHF(mf)
    td.exclude_nlc = not with_nlc
    return _gen_tdhf_operation(td, fock_ao)

def _gen_tdhf_operation(td, fock_ao=None):
    mf = td._scf
    mo_coeff = mf.mo_coeff
    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    nao, nmo = mo_coeff.shape
    n2c = nmo // 2
    occidx = n2c + numpy.where(mo_occ[n2c:] == 1)[0]
    viridx = n2c + numpy.where(mo_occ[n2c:] == 0)[0]
    nocc = len(occidx)
    nvir = len(viridx)
    orbv = mo_coeff[:,viridx]
    orbo = mo_coeff[:,occidx]

    assert fock_ao is None
    e_ia = hdiag = mo_energy[viridx] - mo_energy[occidx,None]
    hdiag = numpy.hstack((hdiag.ravel(), -hdiag.ravel())).real

    mo_coeff = numpy.asarray(numpy.hstack((orbo,orbv)), order='F')
    vresp = td.gen_response(hermi=0)

    def vind(xys):
        xys = numpy.asarray(xys).reshape(-1,2,nocc,nvir)
        xs, ys = xys.transpose(1,0,2,3)
        dms  = lib.einsum('xov,pv,qo->xpq', xs, orbv, orbo.conj())
        dms += lib.einsum('xov,qv,po->xpq', ys, orbv.conj(), orbo)
        v1ao = vresp(dms) # = <mj||nb> Xjb + <mb||nj> Yjb
        # A ~= <aj||ib>, B = <ab||ij>
        # AX + BY
        # = <aj||ib> Xjb + <ab||ij> Yjb
        # = (<mj||nb> Xjb + <mb||nj> Yjb) Cma* Cni
        v1_top = lib.einsum('xpq,qo,pv->xov', v1ao, orbo, orbv.conj())
        # (B*)X + (A*)Y
        # = <ij||ab> Xjb + <ib||aj> Yjb
        # = (<mj||nb> Xjb + <mb||nj> Yjb) Cmi* Cna
        v1_bot = lib.einsum('xpq,po,qv->xov', v1ao, orbo.conj(), orbv)
        v1_top += numpy.einsum('xia,ia->xia', xs, e_ia)  # AX
        v1_bot += numpy.einsum('xia,ia->xia', ys, e_ia)  # (A*)Y

        # (AX, (-A*)Y)
        nz = xys.shape[0]
        hx = numpy.hstack((v1_top.reshape(nz,-1), -v1_bot.reshape(nz,-1)))
        return hx

    return vind, hdiag


class TDHF(TDBase):

    singlet = None

    @lib.with_doc(gen_tdhf_operation.__doc__)
    def gen_vind(self, mf=None):
        assert mf is None or mf is self._scf
        return _gen_tdhf_operation(self)

    def get_init_guess(self, mf, nstates=None, wfnsym=None):
        x0 = TDA.get_init_guess(self, mf, nstates, wfnsym)
        y0 = numpy.zeros_like(x0)
        return numpy.hstack([x0, y0])

    get_precond = rhf.TDHF.get_precond

    def kernel(self, x0=None, nstates=None):
        '''TDHF diagonalization with non-Hermitian eigenvalue solver
        '''
        cpu0 = (logger.process_clock(), logger.perf_counter())
        self.check_sanity()
        self.dump_flags()
        if nstates is None:
            nstates = self.nstates
        else:
            self.nstates = nstates

        log = logger.Logger(self.stdout, self.verbose)

        vind, hdiag = self.gen_vind(self._scf)
        precond = self.get_precond(hdiag)

        def pickeig(w, v, nroots, envs):
            realidx = numpy.where((abs(w.imag) < REAL_EIG_THRESHOLD) &
                                  (w.real > self.positive_eig_threshold))[0]
            # FIXME: Should the amplitudes be real?
            return lib.linalg_helper._eigs_cmplx2real(w, v, realidx,
                                                      real_eigenvectors=False)

        if x0 is None:
            x0 = self.get_init_guess(self._scf, self.nstates)

        self.converged, w, x1 = lr_eig(
            vind, x0, precond, tol_residual=self.conv_tol, lindep=self.lindep,
            nroots=nstates, pick=pickeig, max_cycle=self.max_cycle,
            max_memory=self.max_memory, verbose=log)

        nocc = (self._scf.mo_occ>0).sum()
        nmo = self._scf.mo_occ.size
        n2c = nmo // 2
        nvir = n2c - nocc
        self.e = w
        def norm_xy(z):
            x, y = z.reshape(2,nocc,nvir)
            norm = lib.norm(x)**2 - lib.norm(y)**2
            if norm < 0:
                log.warn('TDDFT amplitudes |X| smaller than |Y|')
            norm = abs(norm)**-.5
            return x*norm, y*norm
        self.xy = [norm_xy(z) for z in x1]

        if self.chkfile:
            lib.chkfile.save(self.chkfile, 'tddft/e', self.e)
            lib.chkfile.save(self.chkfile, 'tddft/xy', self.xy)

        log.timer('TDDFT', *cpu0)
        self._finalize()
        return self.e, self.xy

scf.dhf.DHF.TDA  = scf.dhf.RDHF.TDA  = lib.class_as_method(TDA)
scf.dhf.DHF.TDHF = scf.dhf.RDHF.TDHF = lib.class_as_method(TDHF)

del (OUTPUT_THRESHOLD)



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Cylindrical symmetry

This module is much slower than direct_spin1_symm.

In this implementation, complex orbitals is used to construct the Hamiltonian.
FCI wavefunction (called complex wavefunction here) is solved using the complex
Hamiltonian. For 2D irreps, the real part and the imaginary part of the complex
FCI wavefunction are identical to the Ex and Ey wavefunction obtained from
direct_spin1_symm module. However, any observables from the complex FCI
wavefunction should have an identical one from either Ex or Ey wavefunction
of direct_spin1_symm.
'''

import functools
import ctypes
import numpy
import numpy as np
from pyscf import lib
from pyscf import ao2mo
from pyscf.lib import logger
from pyscf import symm
from pyscf.scf.hf_symm import map_degeneracy
from pyscf.fci import cistring
from pyscf.fci import direct_spin1
from pyscf.fci import direct_spin1_symm
from pyscf.fci.direct_spin1_symm import (
    _sv_associated_det, _strs_angular_momentum, _cyl_sym_orbital_rotation,
    _validate_degen_mapping)
from pyscf.fci import direct_nosym
from pyscf.fci import addons
from pyscf import __config__

libfci = direct_spin1.libfci

def contract_2e(eri, fcivec, norb, nelec, link_index=None, orbsym=None, wfnsym=0):
    if orbsym is None:
        return direct_nosym.contract_2e(eri, fcivec, norb, nelec, link_index)

    neleca, nelecb = direct_spin1._unpack_nelec(nelec)
    link_indexa, link_indexb = direct_nosym._unpack(norb, nelec, link_index)
    na, nlinka = link_indexa.shape[:2]
    nb, nlinkb = link_indexb.shape[:2]

    wfn_momentum = symm.basis.linearmole_irrep2momentum(wfnsym)
    wfnsym_in_d2h = wfnsym % 10
    wfn_ungerade = int(wfnsym_in_d2h >= 4)
    orbsym_d2h = orbsym % 10
    orb_ungerade = orbsym_d2h >= 4
    if np.any(orb_ungerade) or wfn_ungerade:
        max_gerades = 2
    else:
        max_gerades = 1

    orb_l = _get_orb_l(orbsym)
    max_eri_l = abs(orb_l).max() * 2

    strsa = cistring.gen_strings4orblist(range(norb), neleca)
    strsa_l = _strs_angular_momentum(strsa, orbsym)
    max_stra_l = strsa_l.max()
    if neleca == nelecb:
        strsb_l = strsa_l
        max_strb_l = max_stra_l
    else:
        strsb = cistring.gen_strings4orblist(range(norb), nelecb)
        strsb_l = _strs_angular_momentum(strsb, orbsym)
        max_strb_l = strsb_l.max()
    max_momentum = max(max_stra_l, max_strb_l, max_eri_l)

    eri, rank_eri, irrep_eri = reorder_eri(eri, norb, orbsym, max_momentum,
                                           max_gerades)
    eri_ir_dims = eri.ir_dims
    aidx, link_indexa = gen_str_irrep(strsa, orbsym, link_indexa, rank_eri,
                                      irrep_eri, max_momentum, max_gerades)
    nas = nbs = np.array([x.size for x in aidx], dtype=np.int32)
    if neleca == nelecb:
        bidx, link_indexb = aidx, link_indexa
    else:
        bidx, link_indexb = gen_str_irrep(strsb, orbsym, link_indexb, rank_eri,
                                          irrep_eri, max_momentum, max_gerades)
        nbs = np.array([x.size for x in bidx], dtype=np.int32)

    nirreps = (max_momentum * 2 + 1) * max_gerades
    ug_offsets = max_momentum * 2 + 1
    ab_idx = [np.zeros(0, dtype=int)] * nirreps
    for ag in range(max_gerades):
        bg = wfn_ungerade ^ ag
        # abs(al) < max_stra_l and abs(bl := wfn_momentum-al) < max_strb_l
        for al in range(max(-max_stra_l, wfn_momentum-max_strb_l),
                        min( max_stra_l, wfn_momentum+max_strb_l)+1):
            bl = wfn_momentum - al
            stra_ir = al + max_momentum + ag * ug_offsets
            strb_ir = bl + max_momentum + bg * ug_offsets
            ab_idx[stra_ir] = (aidx[stra_ir][:,None] * nb + bidx[strb_ir]).ravel()
    ci_size = np.array([x.size for x in ab_idx], dtype=np.int32)

    if fcivec.size == na * nb:
        ab_idx = np.hstack(ab_idx)
        ci0 = fcivec.ravel()[ab_idx]
    else:
        ci0 = fcivec
    ci1 = np.zeros_like(ci0)

    libfci.FCIcontract_2e_cyl_sym(
        eri.ctypes.data_as(ctypes.c_void_p),
        ci0.ctypes.data_as(ctypes.c_void_p),
        ci1.ctypes.data_as(ctypes.c_void_p),
        eri_ir_dims.ctypes.data_as(ctypes.c_void_p),
        ci_size.ctypes.data_as(ctypes.c_void_p),
        nas.ctypes.data_as(ctypes.c_void_p),
        nbs.ctypes.data_as(ctypes.c_void_p),
        link_indexa.ctypes.data_as(ctypes.c_void_p),
        link_indexb.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(norb), ctypes.c_int(nlinka), ctypes.c_int(nlinkb),
        ctypes.c_int(max_momentum), ctypes.c_int(max_gerades),
        ctypes.c_int(wfn_momentum), ctypes.c_int(wfn_ungerade))

    if fcivec.size == na * nb:
        ci1new = np.zeros(fcivec.size, dtype=fcivec.dtype)
        ci1new[ab_idx] = ci1
        ci1 = ci1new.reshape(fcivec.shape)
    return ci1.view(direct_spin1.FCIvector)

def _get_orb_l(orbsym):
    '''angular momentum for each orbital'''
    orb_l = (orbsym // 10) * 2
    orbsym_d2h = orbsym % 10
    e1_mask = np.isin(orbsym_d2h, (2, 3, 6, 7))
    orb_l[e1_mask] += 1
    ey_mask = np.isin(orbsym_d2h, (1, 3, 4, 6))
    orb_l[ey_mask] *= -1
    return orb_l

def reorder_eri(eri, norb, orbsym, max_momentum, max_gerades):
    eri = eri.reshape(norb,norb,norb,norb)
    # Swap last two indices because they are contracted to the t1 intermediates
    # in FCIcontract_2e_cyl_sym. t1 is generated with swapped orbital indices (a*norb+i).
    eri = eri.transpose(0,1,3,2).reshape(norb**2, norb**2)

    # % 10 to map irrep IDs of Dooh or Coov, etc. to irreps of D2h, C2v
    orbsym_d2h = orbsym % 10
    orb_ungerade = orbsym_d2h >= 4
    nirreps = (max_momentum * 2 + 1) * max_gerades

    # irrep of (ij| pair
    orb_l = _get_orb_l(orbsym)
    ll_prod = (orb_l[:,None] - orb_l).ravel()
    maxll = abs(orb_l).max() * 2

    old_eri_irrep = np.asarray(ll_prod+max_momentum, dtype=np.int32)
    rank_in_irrep = np.empty_like(old_eri_irrep)
    ir_idx_pairs = [None] * nirreps

    if max_gerades == 2:
        ug_offsets = max_momentum * 2 + 1
        ug_prod = (orb_ungerade[:,None] ^ orb_ungerade).ravel()
        old_eri_irrep[ug_prod] += ug_offsets

        # gerade
        idx = np.asarray(np.where((ll_prod == 0) & ~ug_prod)[0], dtype=np.int32)
        ir_idx_pairs[max_momentum] = (idx, idx)
        rank_in_irrep[idx] = np.arange(idx.size, dtype=np.int32)
        # ungerade
        idx = np.asarray(np.where((ll_prod == 0) & ug_prod)[0], dtype=np.int32)
        ir_idx_pairs[max_momentum+ug_offsets] = (idx, idx)
        rank_in_irrep[idx] = np.arange(idx.size, dtype=np.int32)

        for ll in range(1, maxll+1):
            # gerade
            idx_p = np.asarray(np.where((ll_prod == ll) & ~ug_prod)[0], dtype=np.int32)
            idx_m = np.asarray(np.where((ll_prod ==-ll) & ~ug_prod)[0], dtype=np.int32)
            assert idx_p.size == idx_m.size
            if idx_p.size > 0:
                ir_idx_pairs[max_momentum+ll] = (idx_p, idx_p)
                ir_idx_pairs[max_momentum-ll] = (idx_m, idx_m)
                rank_in_irrep[idx_p] = np.arange(idx_p.size, dtype=np.int32)
                rank_in_irrep[idx_m] = np.arange(idx_m.size, dtype=np.int32)
            # ungerade
            idx_p = np.asarray(np.where((ll_prod == ll) & ug_prod)[0], dtype=np.int32)
            idx_m = np.asarray(np.where((ll_prod ==-ll) & ug_prod)[0], dtype=np.int32)
            assert idx_p.size == idx_m.size
            if idx_p.size > 0:
                ir_idx_pairs[max_momentum+ll+ug_offsets] = (idx_p, idx_p)
                ir_idx_pairs[max_momentum-ll+ug_offsets] = (idx_m, idx_m)
                rank_in_irrep[idx_p] = np.arange(idx_p.size, dtype=np.int32)
                rank_in_irrep[idx_m] = np.arange(idx_m.size, dtype=np.int32)
    else:
        idx = np.asarray(np.where(ll_prod == 0)[0], dtype=np.int32)
        ir_idx_pairs[max_momentum] = (idx, idx)
        rank_in_irrep[idx] = np.arange(idx.size, dtype=np.int32)

        for ll in range(1, maxll+1):
            idx_p = np.asarray(np.where(ll_prod == ll)[0], dtype=np.int32)
            idx_m = np.asarray(np.where(ll_prod ==-ll)[0], dtype=np.int32)
            assert idx_p.size == idx_m.size
            if idx_p.size > 0:
                ir_idx_pairs[max_momentum+ll] = (idx_p, idx_p)
                ir_idx_pairs[max_momentum-ll] = (idx_m, idx_m)
                rank_in_irrep[idx_p] = np.arange(idx_p.size, dtype=np.int32)
                rank_in_irrep[idx_m] = np.arange(idx_m.size, dtype=np.int32)

    ir_dims = np.hstack([0 if x is None else x[0].size for x in ir_idx_pairs])
    eri_irs = np.empty((ir_dims**2).sum())
    p1 = 0
    for idx in ir_idx_pairs:
        if idx is not None:
            p0, p1 = p1, p1 + idx[0].size**2
            lib.take_2d(eri, idx[0], idx[1], out=eri_irs[p0:p1])
    eri_irs = lib.tag_array(eri_irs, ir_dims=np.asarray(ir_dims, dtype=np.int32))
    return eri_irs, rank_in_irrep, old_eri_irrep

def argsort_strs_by_irrep(strs, orbsym, max_momentum, max_gerades):
    strs_ls = _strs_angular_momentum(strs, orbsym)
    maxl = abs(strs_ls).max()
    nirreps = (max_momentum * 2 + 1) * max_gerades
    aidx = [np.zeros(0, dtype=np.int32)] * nirreps

    if max_gerades == 2:
        ug_offsets = max_momentum * 2 + 1
        irreps_d2h = direct_spin1_symm._gen_strs_irrep(strs, orbsym)
        strs_ug = irreps_d2h >= 4
        for l in range(-maxl, maxl+1):
            idx = np.where((strs_ls == l) & ~strs_ug)[0]
            aidx[max_momentum+l] = idx
            idx = np.where((strs_ls == l) & strs_ug)[0]
            aidx[max_momentum+l+ug_offsets] = idx
    else:
        for l in range(-maxl, maxl+1):
            idx = np.where(strs_ls == l)[0]
            aidx[max_momentum+l] = idx
    return aidx

def gen_str_irrep(strs, orbsym, link_index, rank_eri, irrep_eri, max_momentum,
                  max_gerades):
    aidx = argsort_strs_by_irrep(strs, orbsym, max_momentum, max_gerades)
    na = len(strs)
    rank = np.zeros(na, dtype=np.int32)
    for idx in aidx:
        if idx.size > 0:
            rank[idx] = np.arange(idx.size, dtype=np.int32)

    link_index = link_index.copy()
    norb = orbsym.size
    ai_addr = link_index[:,:,0] * norb + link_index[:,:,1]
    link_index[:,:,0] = rank_eri[ai_addr]
    link_index[:,:,1] = irrep_eri[ai_addr]
    link_index[:,:,2] = rank[link_index[:,:,2]]

    link_index = link_index.take(np.hstack(aidx), axis=0)
    return aidx, link_index

def get_init_guess(norb, nelec, nroots, hdiag, orbsym, wfnsym=0,
                   sym_allowed_idx=None):
    neleca, nelecb = direct_spin1._unpack_nelec(nelec)
    strsa = strsb = cistring.gen_strings4orblist(range(norb), neleca)
    if neleca != nelecb:
        strsb = cistring.gen_strings4orblist(range(norb), nelecb)
    na = len(strsa)
    nb = len(strsb)
    degen = orbsym.degen_mapping

    if sym_allowed_idx is None:
        sym_allowed_idx = sym_allowed_indices(nelec, orbsym, wfnsym)
    s_idx = np.hstack(sym_allowed_idx)
    idx_a, idx_b = divmod(s_idx, nb)
    if hdiag.size == na*nb:
        hdiag = hdiag[s_idx]
    civec_size = hdiag.size

    if neleca == nelecb and na == nb:
        idx = np.arange(idx_a.size)[idx_a >= idx_b]
        idx_a = idx_a[idx]
        idx_b = idx_b[idx]
        hdiag = hdiag[idx]

    ci0 = []
    for k in hdiag.argsort():
        addra, addrb = idx_a[k], idx_b[k]
        x = np.zeros(civec_size)
        x[s_idx==addra*nb+addrb] = 1.
        if wfnsym in (0, 1, 4, 5):
            addra1, sign_a = _sv_associated_det(strsa[addra], degen)
            addrb1, sign_b = _sv_associated_det(strsb[addrb], degen)
            # If (E+) and (E-) are associated determinants
            # (E+)(E-') + (E-)(E+') => A1
            # (E+)(E-') - (E-)(E+') => A2
            if wfnsym in (0, 5):  # A1g, A1u
                # ensure <|sigma_v|> = 1
                x[s_idx==addra1*nb+addrb1] += sign_a * sign_b
            elif wfnsym in (1, 4):  # A2g, A2u
                # ensure <|sigma_v|> = -1
                x[s_idx==addra1*nb+addrb1] -= sign_a * sign_b

        norm = np.linalg.norm(x)
        if norm < 1e-3:
            continue
        x *= 1./norm
        ci0.append(x.view(direct_spin1.FCIvector))
        if len(ci0) >= nroots:
            break

    if len(ci0) == 0:
        raise RuntimeError(f'Initial guess for symmetry {wfnsym} not found')
    return ci0


def _guess_wfnsym(civec, strsa, strsb, orbsym):
    degen_mapping = orbsym.degen_mapping
    idx = abs(civec).argmax()
    na = strsa.size
    nb = strsb.size
    civec = civec.reshape(na,nb)
    addra = idx // nb
    addrb = idx % nb
    addra1, sign_a = _sv_associated_det(strsa[addra], degen_mapping)
    addrb1, sign_b = _sv_associated_det(strsb[addrb], degen_mapping)

    airreps_d2h = direct_spin1_symm._gen_strs_irrep(strsa[[addra]], orbsym)
    birreps_d2h = direct_spin1_symm._gen_strs_irrep(strsb[[addrb]], orbsym)
    a_ls = _strs_angular_momentum(strsa[[addra]], orbsym)
    b_ls = _strs_angular_momentum(strsb[[addrb]], orbsym)
    a_ungerade = airreps_d2h >= 4
    b_ungerade = birreps_d2h >= 4
    wfn_ungerade = a_ungerade[0] ^ b_ungerade[0]
    wfn_momentum = a_ls[0] + b_ls[0]

    if wfn_momentum == 0:
        # For A1g and A1u, CI coefficient and its sigma_v associated one have
        # the same sign
        if sign_a*sign_b * civec[addra,addrb] * civec[addra1,addrb1] > 1e-6: # A1
            if wfn_ungerade:
                wfnsym = 5
            else:
                wfnsym = 0
        else:
            # For A2g and A2u, CI coefficient and its sigma_v associated one
            # have opposite signs
            if wfn_ungerade:
                wfnsym = 4
            else:
                wfnsym = 1

    elif wfn_momentum % 2 == 1:
        if wfn_momentum > 0:  # Ex
            if wfn_ungerade:
                wfnsym = 7
            else:
                wfnsym = 2
        else:  # Ey
            if wfn_ungerade:
                wfnsym = 6
            else:
                wfnsym = 3
    else:
        if wfn_momentum > 0:  # Ex
            if wfn_ungerade:
                wfnsym = 5
            else:
                wfnsym = 0
        else:  # Ey
            if wfn_ungerade:
                wfnsym = 4
            else:
                wfnsym = 1

    wfnsym += (abs(wfn_momentum) // 2) * 10
    return wfnsym

def guess_wfnsym(solver, norb, nelec, fcivec=None, orbsym=None, wfnsym=None, **kwargs):
    '''
    Guess point group symmetry of the FCI wavefunction.  If fcivec is
    given, the symmetry of fcivec is used.  Otherwise the symmetry is
    same to the HF determinant.
    '''
    if orbsym is None:
        orbsym = solver.orbsym

    verbose = kwargs.get('verbose', None)
    log = logger.new_logger(solver, verbose)

    neleca, nelecb = nelec = direct_spin1._unpack_nelec(nelec, solver.spin)
    if fcivec is None or not hasattr(orbsym, 'degen_mapping'):
        # guess wfnsym if initial guess is not given
        wfnsym = direct_spin1_symm._id_wfnsym(solver, norb, nelec, orbsym, wfnsym)
        log.debug('Guessing CI wfn symmetry = %s', wfnsym)

    else:
        strsa = strsb = cistring.gen_strings4orblist(range(norb), neleca)
        if neleca != nelecb:
            strsb = cistring.gen_strings4orblist(range(norb), nelecb)

        if not isinstance(fcivec, np.ndarray) or fcivec.ndim > 2:
            fcivec = fcivec[0]
        wfnsym1 = _guess_wfnsym(fcivec, strsa, strsb, orbsym)

        if (wfnsym is not None and
            wfnsym1 != direct_spin1_symm._id_wfnsym(solver, norb, nelec, orbsym, wfnsym)):
            raise RuntimeError(f'Input wfnsym {wfnsym} is not consistent with '
                               f'fcivec symmetry {wfnsym1}')
        wfnsym = wfnsym1
    return wfnsym

def sym_allowed_indices(nelec, orbsym, wfnsym):
    '''Indices of symmetry allowed determinants for each irrep'''
    norb = orbsym.size
    neleca, nelecb = nelec
    strsa = strsb = cistring.gen_strings4orblist(range(norb), neleca)
    strsa_l = _strs_angular_momentum(strsa, orbsym)
    max_stra_l = max_strb_l = strsa_l.max()
    if neleca != nelecb:
        strsb = cistring.gen_strings4orblist(range(norb), nelecb)
        strsb_l = _strs_angular_momentum(strsb, orbsym)
        max_strb_l = strsb_l.max()
    nb = len(strsb)

    wfn_momentum = symm.basis.linearmole_irrep2momentum(wfnsym)
    wfnsym_in_d2h = wfnsym % 10
    wfn_ungerade = wfnsym_in_d2h >= 4
    orbsym_d2h = orbsym % 10
    orb_ungerade = orbsym_d2h >= 4
    if np.any(orb_ungerade) or wfn_ungerade:
        max_gerades = 2
    else:
        max_gerades = 1
    orb_l = _get_orb_l(orbsym)
    max_eri_l = abs(orb_l).max() * 2
    max_momentum = max(max_stra_l, max_strb_l, max_eri_l)

    aidx = bidx = argsort_strs_by_irrep(strsa, orbsym, max_momentum, max_gerades)
    if neleca != nelecb:
        bidx = argsort_strs_by_irrep(strsb, orbsym, max_momentum, max_gerades)

    nirreps = (max_momentum * 2 + 1) * max_gerades
    ug_offsets = max_momentum * 2 + 1
    ab_idx = [np.zeros(0, dtype=int)] * nirreps
    for ag in range(max_gerades):
        bg = wfn_ungerade ^ ag
        # abs(al) < max_stra_l and abs(bl := wfn_momentum-al) < max_strb_l
        for al in range(max(-max_stra_l, wfn_momentum-max_strb_l),
                        min( max_stra_l, wfn_momentum+max_strb_l)+1):
            bl = wfn_momentum - al
            stra_ir = al + max_momentum + ag * ug_offsets
            strb_ir = bl + max_momentum + bg * ug_offsets
            ab_idx[stra_ir] = (aidx[stra_ir][:,None] * nb + bidx[strb_ir]).ravel()
    return ab_idx

def _dm_wrapper(fn_rdm):
    def transform(dm, u):
        if dm.ndim == 2:
            dm = u.conj().T.dot(dm).dot(u)
        else:
            dm = lib.einsum('pqrs,pi,qj,rk,sl->ijkl', dm, u.conj(), u, u.conj(), u)
        return dm.real.copy()

    @functools.wraps(fn_rdm)
    def make_dm(self, fcivec, norb, nelec, *args, **kwargs):
        nelec = direct_spin1._unpack_nelec(nelec, self.spin)
        dms = fn_rdm(fcivec, norb, nelec, *args, **kwargs)
        orbsym = self.orbsym
        degen_mapping = self.orbsym.degen_mapping
        u = _cyl_sym_orbital_rotation(orbsym, degen_mapping)
        if isinstance(dms, np.ndarray):
            return transform(dms, u)
        else:
            return [transform(dm, u) for dm in dms]
    return make_dm


class FCISolver(direct_spin1_symm.FCISolver):

    def contract_1e(self, f1e, fcivec, norb, nelec, link_index=None, **kwargs):
        raise NotImplementedError

    def contract_2e(self, eri, fcivec, norb, nelec, link_index=None,
                    orbsym=None, wfnsym=None, **kwargs):
        if orbsym is None: orbsym = self.orbsym
        if wfnsym is None:
            wfnsym = direct_spin1_symm._id_wfnsym(self, norb, nelec, orbsym, self.wfnsym)
        return contract_2e(eri, fcivec, norb, nelec, link_index, orbsym, wfnsym)

    def get_init_guess(self, norb, nelec, nroots, hdiag, orbsym=None, wfnsym=None):
        if orbsym is None:
            orbsym = self.orbsym
        if wfnsym is None:
            wfnsym = direct_spin1_symm._id_wfnsym(self, norb, nelec, orbsym, self.wfnsym)
        return get_init_guess(norb, nelec, nroots, hdiag, orbsym, wfnsym,
                              self.sym_allowed_idx)

    def pspace(self, h1e, eri, norb, nelec, hdiag, np=400):
        nelec = direct_spin1._unpack_nelec(nelec, self.spin)
        na = cistring.num_strings(norb, nelec[0])
        nb = cistring.num_strings(norb, nelec[1])
        s_idx = numpy.hstack(self.sym_allowed_idx)
        # Screen symmetry forbidden elements
        hdiag, hdiag0 = numpy.empty(na*nb), hdiag.ravel()
        hdiag[:] = 1e99
        if hdiag0.size == s_idx.size:
            hdiag[s_idx] = hdiag0
        else:
            hdiag[s_idx] = hdiag0[s_idx]

        np = min(np, s_idx.size)
        addr0, h = direct_spin1.pspace(h1e, eri, norb, nelec, hdiag, np)

        # mapping the address in (na,nb) civec to address in sym-allowed civec
        s_idx_allowed = numpy.where(numpy.isin(s_idx, addr0))[0]
        addr1 = s_idx[s_idx_allowed]
        new_idx = numpy.empty_like(s_idx_allowed)
        new_idx[addr0.argsort()] = addr1.argsort()
        addr = s_idx_allowed[new_idx]
        return addr, h

    absorb_h1e = direct_nosym.FCISolver.absorb_h1e
    make_hdiag = direct_spin1_symm.FCISolver.make_hdiag
    guess_wfnsym = guess_wfnsym

    make_rdm1 = _dm_wrapper(direct_spin1.make_rdm1)
    make_rdm1s = _dm_wrapper(direct_spin1.make_rdm1s)
    make_rdm12 = _dm_wrapper(direct_spin1.make_rdm12)
    make_rdm12s = _dm_wrapper(direct_spin1.make_rdm12s)
    trans_rdm1 = _dm_wrapper(direct_spin1.trans_rdm1)
    trans_rdm1s = _dm_wrapper(direct_spin1.trans_rdm1s)
    trans_rdm12 = _dm_wrapper(direct_spin1.trans_rdm12)
    trans_rdm12s = _dm_wrapper(direct_spin1.trans_rdm12s)

    def kernel(self, h1e, eri, norb, nelec, ci0=None,
               tol=None, lindep=None, max_cycle=None, max_space=None,
               nroots=None, davidson_only=None, pspace_size=None,
               orbsym=None, wfnsym=None, ecore=0, **kwargs):
        if nroots is None: nroots = self.nroots
        if orbsym is None: orbsym = self.orbsym
        if wfnsym is None: wfnsym = self.wfnsym
        if self.verbose >= logger.WARN:
            if 'verbose' not in kwargs:
                kwargs['verbose'] = self.verbose
            self.check_sanity()
        self.norb = norb
        self.nelec = nelec = direct_spin1._unpack_nelec(nelec, self.spin)

        if not hasattr(orbsym, 'degen_mapping'):
            degen_mapping = map_degeneracy(h1e.diagonal(), orbsym)
            orbsym = lib.tag_array(orbsym, degen_mapping=degen_mapping)
        if not _validate_degen_mapping(orbsym.degen_mapping, norb):
            raise lib.exceptions.PointGroupSymmetryError(
                'Incomplete 2D-irrep orbitals for cylindrical symmetry.\n'
                f'orbsym = {orbsym}.')

        u = _cyl_sym_orbital_rotation(orbsym, orbsym.degen_mapping)
        h1e = u.dot(h1e).dot(u.conj().T)
        eri = ao2mo.restore(1, eri, norb)
        eri = lib.einsum('pqrs,ip,jq,kr,ls->ijkl', eri, u, u.conj(), u, u.conj())
        assert abs(h1e.imag).max() < 1e-12, 'Cylindrical symmetry broken'
        assert abs(eri.imag).max() < 1e-12, 'Cylindrical symmetry broken'
        h1e = h1e.real.copy()
        # Note: eri is real but it does not have the permutation relation
        # (ij|kl) = (ji|kl) = (ij|lk) = (ji|lk)
        # The nosym version fci contraction is required
        eri = eri.real.copy()

        wfnsym_ir = self.guess_wfnsym(norb, nelec, ci0, orbsym, wfnsym, **kwargs)
        if wfnsym_ir in (1, 4):
            # sym_allowed_idx does not distinguish A2g and A2u
            davidson_only = True
        self.sym_allowed_idx = sym_allowed_indices(nelec, orbsym, wfnsym_ir)
        self.orbsym = orbsym
        logger.debug(self, 'Num symmetry allowed elements %d',
                     sum([x.size for x in self.sym_allowed_idx]))

        neleca, nelecb = direct_spin1._unpack_nelec(nelec)
        link_indexa = cistring.gen_linkstr_index(range(norb), neleca)
        if neleca == nelecb:
            link_indexb = link_indexa
        else:
            link_indexb = cistring.gen_linkstr_index(range(norb), nelecb)

        with lib.temporary_env(self, wfnsym=wfnsym_ir):
            e, c = direct_spin1.kernel_ms1(self, h1e, eri, norb, nelec, ci0,
                                           (link_indexa,link_indexb),
                                           tol, lindep, max_cycle, max_space,
                                           nroots, davidson_only, pspace_size,
                                           ecore=ecore, **kwargs)

        na = link_indexa.shape[0]
        nb = link_indexb.shape[0]
        s_idx = np.hstack(self.sym_allowed_idx)
        if nroots > 1:
            c, c_raw = [], c
            for vec in c_raw:
                c1 = np.zeros(na*nb)
                c1[s_idx] = vec.T
                c.append(c1)
        else:
            c1 = np.zeros(na*nb)
            c1[s_idx] = c
            c = c1.reshape(na, nb).view(direct_spin1.FCIvector)

        self.eci, self.ci = e, c
        return e, c

FCI = FCISolver



#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xiaojie Wu <wxj6000@gmail.com>
#

'''
gradient of dispersion correction for HF and DFT
'''

import numpy as np
from pyscf.lib import logger
from pyscf.scf.dispersion import check_disp, parse_disp

def get_dispersion(mf_grad, disp=None, with_3body=None, verbose=None):
    '''gradient of DFTD3/DFTD4 dispersion correction'''
    mf = mf_grad.base
    mol = mf.mol
    disp_version = check_disp(mf, disp)
    if not disp_version:
        return np.zeros([mol.natm,3])

    try:
        from pyscf.dispersion import dftd3, dftd4
    except ImportError:
        print('dftd3 and dftd4 not available. Install them with `pip install pyscf-dispersion`')
        raise

    method = getattr(mf, 'xc', 'hf')
    method, _, disp_with_3body = parse_disp(method)

    if with_3body is not None:
        with_3body = disp_with_3body

    if disp_version[:2].upper() == 'D3':
        logger.info(mf, "Calc dispersion correction with DFTD3.")
        logger.info(mf, f"Parameters: xc={method}, version={disp_version}, atm={with_3body}")
        d3_model = dftd3.DFTD3Dispersion(mol, xc=method, version=disp_version, atm=with_3body)
        res = d3_model.get_dispersion(grad=True)
        g_d3 = res.get('gradient')
        return g_d3
    elif disp_version[:2].upper() == 'D4':
        logger.info(mf, "Calc dispersion correction with DFTD4.")
        logger.info(mf, f"Parameters: xc={method}, atm={with_3body}")
        d4_model = dftd4.DFTD4Dispersion(mol, xc=method, atm=with_3body)
        res = d4_model.get_dispersion(grad=True)
        g_d4 = res.get('gradient')
        return g_d4
    else:
        raise RuntimeError(f'dispersion correction: {disp_version} is not supported.')

# Inject to Gradient
from pyscf import grad
grad.rhf.GradientsBase.get_dispersion = get_dispersion



#!/usr/bin/env python
# Copyright 2020-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Xing Zhang <zhangxing.nju@gmail.com>
#

import sys
import copy
from functools import reduce
import numpy as np
from numpy.linalg import inv, det
from pyscf import lib
from pyscf.lib import logger
from pyscf.symm import param
from pyscf.symm.Dmatrix import Dmatrix, get_euler_angles
from pyscf.pbc.symm import space_group
from pyscf.pbc.symm.space_group import SYMPREC, XYZ
from pyscf.pbc.tools import pbc as pbctools

def get_Dmat(op, l):
    '''
    Get Wigner D-matrix

    Args:
        op : (3,3) ndarray
            rotation operator in (x,y,z) system
        l : int
            angular momentum
    '''
    fac = 1
    det_op = det(op)
    if det_op < 0:
        # improper rotation has |R| = -1
        assert abs(det_op + 1) < 1e-9
        op = -1 * op
        fac = (-1) ** l

    c1 = XYZ
    c2 = np.dot(op, c1.T).T
    alpha, beta, gamma = get_euler_angles(c1, c2)
    D = fac * Dmatrix(l, alpha, beta, gamma, reorder_p=True)
    return D.round(15)

def get_Dmat_cart(op,l_max):
    pp = get_Dmat(op, 1)
    Ds = [np.ones((1,1))]
    for l in range(1, l_max+1):
        # All possible x,y,z combinations
        cidx = np.sort(lib.cartesian_prod([(0, 1, 2)] * l), axis=1)

        addr = 0
        affine = np.ones((1,1))
        for i in range(l):
            nd = affine.shape[0] * 3
            affine = np.einsum('ik,jl->ijkl', affine, pp).reshape(nd, nd)
            addr = addr * 3 + cidx[:,i]

        uniq_addr, rev_addr = np.unique(addr, return_inverse=True)
        ncart = (l + 1) * (l + 2) // 2
        assert ncart == uniq_addr.size
        trans = np.zeros((ncart,ncart))
        for i, k in enumerate(rev_addr):
            trans[k] += affine[i,uniq_addr]
        Ds.append(trans)
    return Ds

def make_Dmats(cell, ops, l_max=None):
    '''
    Computes < m | R | m' >
    '''
    if l_max is None:
        l_max = np.max(cell._bas[:,1])
    else:
        l_max = max(l_max, np.max(cell._bas[:,1]))

    Dmats = []
    for op in ops:
        if not cell.cart:
            Dmats.append([get_Dmat(op, l) for l in range(l_max+1)])
        else:
            Dmats.append(get_Dmat_cart(op, l_max))
    return Dmats, l_max

def check_mesh_symmetry(cell, ops, mesh=None, tol=SYMPREC,
                        return_mesh=False):
    if mesh is None:
        mesh = cell.mesh
    ft = []
    rm_list = []
    for i, op in enumerate(ops):
        if not op.trans_is_zero:
            ft.append(op.trans)
            tmp = op.trans * np.asarray(mesh)
            if (abs(tmp - tmp.round()) > tol).any():
                rm_list.append(i)

    if len(rm_list) == 0:
        mesh1 = mesh
    else:
        ft = np.reshape(np.asarray(ft), (-1,3))
        mesh1 = copy.deepcopy(mesh)
        for x in range(3):
            while True:
                tmp = ft[:,x] * mesh1[x]
                if (abs(tmp - tmp.round()) > tol).any():
                    mesh1[x] = mesh1[x] + 1
                else:
                    break

        if not return_mesh:
            logger.warn(cell, 'Input mesh %s has lower symmetry than the lattice.\n'
                        'Some of the symmetry operations will be removed.\n'
                        'Recommended mesh is %s.', mesh, mesh1)
    if return_mesh:
        return rm_list, mesh1
    else:
        return rm_list

class Symmetry():
    '''
    Symmetry info of a crystal.

    Attributes:
        cell : :class:`Cell` object

        spacegroup : :class:`SpaceGroup` object

        symmorphic : bool
            Whether space group is symmorphic
        has_inversion : bool
            Whether space group contains inversion operation
        ops : list of :class:`SPGElement` object
            Symmetry operators (may be a subset of the operators in the space group)
        nop : int
            Length of `ops`.
        Dmats : list of 2d arrays
            Wigner D-matrices
        l_max : int
            Maximum angular momentum considered in `Dmats`
    '''
    def __init__(self, cell):
        self.cell = cell
        self.spacegroup = None
        self.symmorphic = True
        self.ops = [space_group.SPGElement(),]
        self.nop = len(self.ops)
        self.has_inversion = False
        self.Dmats = None
        self.l_max = None
        self._built = False

    def build(self, space_group_symmetry=True, symmorphic=True,
              check_mesh_symmetry=True, *args, **kwargs):
        cell = self.cell
        if cell is None:
            self._built = True
            return self

        if not space_group_symmetry:
            self.ops = [space_group.SPGElement(),]
        else:
            if not cell._built:
                sys.stderr.write('Warning: %s must be initialized before calling Symmetry.\n'
                                 'Initialize %s in %s\n' % (cell, cell, self))
                cell.build()

            self.spacegroup = space_group.SpaceGroup(cell).build(dump_info=False)
            self.symmorphic = symmorphic
            if cell.dimension < 3:
                if not self.symmorphic:
                    sys.stderr.write('Warning: setting symmorphic=True for low-dimensional system.\n')
                    self.symmorphic = True

            ops = self.spacegroup.ops
            if self.symmorphic:
                self.ops = [op for op in ops if op.trans_is_zero]
            elif check_mesh_symmetry:
                rm_list = self.check_mesh_symmetry(ops=ops)
                self.ops = [op for i, op in enumerate(ops) if i not in rm_list]
            else:
                self.ops = ops

        self.nop = len(self.ops)
        self.has_inversion = any(op.rot_is_inversion for op in self.ops)

        l_max = None
        if 'auxcell' in kwargs:
            auxcell = kwargs['auxcell']
            if getattr(auxcell, '_bas', None) is not None:
                l_max = np.max(auxcell._bas[:,1])
        op_rot = [op.a2r(self.cell).rot for op in self.ops]
        self.Dmats, self.l_max = make_Dmats(self.cell, op_rot, l_max)
        self._built = True
        return self

    def check_mesh_symmetry(self, cell=None, ops=None, mesh=None,
                            tol=SYMPREC, return_mesh=False):
        if cell is None:
            cell = self.cell
        if ops is None:
            ops = self.ops
        return check_mesh_symmetry(cell, ops, mesh, tol, return_mesh)

    def dump_info(self):
        self.spacegroup.dump_info(ops=self.ops)


def _get_phase(cell, op, kpt_scaled, ignore_phase=False, tol=SYMPREC):
    kpt_scaled = op.a2b(cell).dot_rot(kpt_scaled)
    coords_scaled = cell.get_scaled_atom_coords().reshape(-1,3)
    natm = coords_scaled.shape[0]
    phase = np.ones((natm,), dtype=np.complex128)
    atm_map = np.arange(natm)
    coords0 = pbctools.round_to_cell0(coords_scaled, tol=tol)
    for iatm in range(natm):
        r = coords_scaled[iatm]
        op_dot_r = op.dot_rot(r) + op.trans
        op_dot_r_0 = pbctools.round_to_cell0(op_dot_r, tol=tol)
        equiv_atm = np.where(abs(op_dot_r_0 - coords0).sum(axis=1) < tol)[0]
        assert len(equiv_atm) == 1
        equiv_atm = equiv_atm[0]
        atm_map[iatm] = equiv_atm
        Lshift = coords_scaled[equiv_atm] - op_dot_r
        # Lshift is a lattice vector
        assert abs(Lshift - Lshift.round()).sum() < tol
        # remove numerical noise, important for symmetry adaptation
        Lshift = Lshift.round()
        if not ignore_phase:
            phase[iatm] = np.exp(1j * np.dot(kpt_scaled, Lshift) * 2.0 * np.pi)
    return atm_map, phase

def _get_rotation_mat(cell, kpt_scaled_ibz, mo_coeff_or_dm, op, Dmats,
                      ignore_phase=False, tol=SYMPREC):
    atm_map, phases = _get_phase(cell, op, kpt_scaled_ibz, ignore_phase, tol)

    dim = mo_coeff_or_dm.shape[0]
    mat = np.zeros([dim, dim], dtype=np.complex128)
    aoslice = cell.aoslice_by_atom()
    for iatm in range(cell.natm):
        jatm = atm_map[iatm]
        if iatm != jatm:
            #sanity check
            nao_i = aoslice[iatm][3] - aoslice[iatm][2]
            nao_j = aoslice[jatm][3] - aoslice[jatm][2]
            assert(nao_i == nao_j)
            nshl_i = aoslice[iatm][1] - aoslice[iatm][0]
            nshl_j = aoslice[jatm][1] - aoslice[jatm][0]
            assert(nshl_i == nshl_j)
            for ishl in range(nshl_i):
                shl_i = ishl + aoslice[iatm][0]
                shl_j = ishl + aoslice[jatm][0]
                l_i = cell._bas[shl_i,1]
                l_j = cell._bas[shl_j,1]
                assert(l_i == l_j)
        phase = phases[iatm]
        ao_off_i = aoslice[iatm][2]
        ao_off_j = aoslice[jatm][2]
        shlid_0 = aoslice[iatm][0]
        shlid_1 = aoslice[iatm][1]
        for ishl in range(shlid_0, shlid_1):
            l = cell.bas_angular(ishl)
            Dmat = Dmats[l] * phase
            if not cell.cart:
                nao = 2 * l + 1
            else:
                nao = (l+1) * (l+2) // 2
            nc = cell.bas_nctr(ishl)
            for _ in range(nc):
                mat[ao_off_j:ao_off_j+nao, ao_off_i:ao_off_i+nao] = Dmat
                ao_off_i += nao
                ao_off_j += nao
        assert ao_off_i == aoslice[iatm][3]
        assert ao_off_j == aoslice[jatm][3]
    return mat

def transform_mo_coeff(cell, kpt_scaled, mo_coeff, op, Dmats):
    '''
    Get MO coefficients at a symmetry-related k-point

    Args:
        cell : :class:`Cell` object

        kpt_scaled : (3,) array
            scaled k-point
        mo_coeff : (nao, nmo) array
            MO coefficients at the input k-point
        op : :class:`SPGElement` object
            Space group operation that connects the two k-points
        Dmats: list of arrays
            Wigner D-matrices for op

    Returns:
        MO coefficients at the symmetry-related k-point
    '''
    mat = _get_rotation_mat(cell, kpt_scaled, mo_coeff, op, Dmats)
    return np.dot(mat, mo_coeff)

def transform_dm(cell, kpt_scaled, dm, op, Dmats):
    '''
    Get density matrix for a symmetry-related k-point
    '''
    mat = _get_rotation_mat(cell, kpt_scaled, dm, op, Dmats)
    return reduce(np.dot, (mat, dm, mat.T.conj()))

def transform_1e_operator(cell, kpt_scaled, fock, op, Dmats):
    '''
    Get 1-electron operator for a symmetry-related k-point
    '''
    mat = _get_rotation_mat(cell, kpt_scaled, fock, op, Dmats)
    return reduce(np.dot, (mat, fock, mat.T.conj()))

def make_rot_loc(l_max, key):
    l = np.arange(l_max+1)
    if 'cart' in key:
        dims = ((l+1)*(l+2)//2)**2
    elif 'sph' in key:
        dims = (l*2+1)**2
    else:  # spinor
        raise NotImplementedError

    rot_loc = np.empty(len(dims)+1, dtype=np.int32)
    rot_loc[0] = 0
    dims.cumsum(dtype=np.int32, out=rot_loc[1:])
    return rot_loc

def is_eye(op):
    raise NotImplementedError

def is_inversion(op):
    raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Transform XC functional derivatives between different representations
'''

import math
import itertools
from functools import lru_cache
import ctypes
import numpy as np
from pyscf import lib

libdft = lib.load_library('libdft')


def transform_vxc(rho, vxc, xctype, spin=0):
    r'''
    Transform libxc functional derivatives to the derivative tensor of
    parameters in rho:
        [[density_a, (nabla_x)_a, (nabla_y)_a, (nabla_z)_a, tau_a],
         [density_b, (nabla_x)_b, (nabla_y)_b, (nabla_z)_b, tau_b]].
    The output tensor has the shape:
        * spin polarized
            LDA : [2,1,N]
            GGA : [2,4,N]
            MGGA: [2,5,N]
        * spin unpolarized
            LDA : [1,N]
            GGA : [4,N]
            MGGA: [5,N]
    '''
    rho = np.asarray(rho, order='C')
    if xctype == 'GGA':
        order = 1
        nvar = 4
        fr = vxc[0].T
        fg = vxc[1].T
    elif xctype == 'MGGA':
        order = 2
        nvar = 5
        fr = vxc[0].T
        fg = vxc[1].T
        ft = vxc[3].T
    else:  # LDA
        order = 0
        nvar = 1
        fr = vxc[0].T

    ngrids = rho.shape[-1]
    if spin == 1:
        if order == 0:
            vp = fr.reshape(2, nvar, ngrids)
        else:
            vp = np.empty((2,nvar, ngrids))
            vp[:,0] = fr
            #:vp[:,1:4] = np.einsum('abg,bxg->axg', _stack_fg(fg), rho[:,1:4])
            vp[:,1:4] = _stack_fg(fg, rho=rho)
        if order > 1:
            vp[:,4] = ft
    else:
        if order == 0:
            vp = fr.reshape(nvar, ngrids)
        else:
            vp = np.empty((nvar, ngrids))
            vp[0] = fr
            vp[1:4] = 2 * fg * rho[1:4]
        if order > 1:
            vp[4] = ft
    return vp

def transform_fxc(rho, vxc, fxc, xctype, spin=0):
    r'''
    Transform libxc functional derivatives to the derivative tensor of
    parameters in rho:
        [[density_a, (nabla_x)_a, (nabla_y)_a, (nabla_z)_a, tau_a],
         [density_b, (nabla_x)_b, (nabla_y)_b, (nabla_z)_b, tau_b]].
    The output tensor has the shape:
        * spin polarized
            LDA : [2,1,2,1,N]
            GGA : [2,4,2,4,N]
            MGGA: [2,5,2,5,N]
        * spin unpolarized
            LDA : [1,1,N]
            GGA : [4,4,N]
            MGGA: [5,5,N]
    '''
    rho = np.asarray(rho, order='C')
    if xctype == 'GGA':
        order = 1
        nvar = 4
        fg = vxc[1].T
        frr = fxc[0].T
        frg = fxc[1].T
        fgg = fxc[2].T
    elif xctype == 'MGGA':
        order = 2
        nvar = 5
        fg = vxc[1].T
        frr, frg, fgg, ftt, frt, fgt = [fxc[i].T for i in [0, 1, 2, 4, 6, 9]]
    else:  # LDA
        order = 0
        nvar = 1
        frr = fxc[0].T

    ngrids = rho.shape[-1]
    if spin == 1:
        if order == 0:
            vp = _stack_frr(frr).reshape(2,nvar, 2,nvar, ngrids).transpose(1,3,0,2,4)
        else:
            vp = np.empty((2,nvar, 2,nvar, ngrids)).transpose(1,3,0,2,4)
            vp[0,0] = _stack_frr(frr)
            i3 = np.arange(3)
            #:qgg = _stack_fgg(fgg)
            #:qgg = np.einsum('abcdg,axg->xbcdg', qgg, rho[:,1:4])
            #:qgg = np.einsum('xbcdg,cyg->xybdg', qgg, rho[:,1:4])
            qgg = _stack_fgg(fgg, rho=rho).transpose(1,3,0,2,4)
            qgg[i3,i3] += _stack_fg(fg)
            vp[1:4,1:4] = qgg

            frg = frg.reshape(2,3,ngrids)
            #:qrg = _stack_fg(frg, axis=1)
            #:qrg = np.einsum('rabg,axg->xrbg', qrg, rho[:,1:4])
            qrg = _stack_fg(frg, axis=1, rho=rho).transpose(2,0,1,3)
            vp[0,1:4] = qrg
            vp[1:4,0] = qrg.transpose(0,2,1,3)

        if order > 1:
            fgt = fgt.reshape(3,2,ngrids)
            #:qgt = _stack_fg(fgt, axis=0)
            #:qgt = np.einsum('abrg,axg->xbrg', qgt, rho[:,1:4])
            qgt = _stack_fg(fgt, axis=0, rho=rho).transpose(1,0,2,3)
            vp[1:4,4] = qgt
            vp[4,1:4] = qgt.transpose(0,2,1,3)

            qrt = frt.reshape(2,2,ngrids)
            vp[0,4] = qrt
            vp[4,0] = qrt.transpose(1,0,2)

            vp[4,4] = _stack_frr(ftt)

        vp = vp.transpose(2,0,3,1,4)

    else:
        if order == 0:
            vp = frr.reshape(nvar, nvar, ngrids)
        else:
            vp = np.empty((nvar, nvar, ngrids))
            vp[0,0] = frr
            i3 = np.arange(3)
            qgg = 4 * fgg * rho[1:4] * rho[1:4,None]
            qgg[i3,i3] += fg * 2
            vp[1:4,1:4] = qgg
            vp[0,1:4] = vp[1:4,0] = 2 * frg * rho[1:4]
        if order > 1:
            vp[4,1:4] = vp[1:4,4] = 2 * fgt * rho[1:4]
            vp[0,4] = frt
            vp[4,0] = frt
            vp[4,4] = ftt
    return vp

def transform_kxc(rho, fxc, kxc, xctype, spin=0):
    r'''
    Transform libxc functional derivatives to the derivative tensor of
    parameters in rho:
        [[density_a, (nabla_x)_a, (nabla_y)_a, (nabla_z)_a, tau_a],
         [density_b, (nabla_x)_b, (nabla_y)_b, (nabla_z)_b, tau_b]].
    The output tensor has the shape:
        * spin polarized
            LDA : [2,1,2,1,2,1,N]
            GGA : [2,4,2,4,2,4,N]
            MGGA: [2,5,2,5,2,5,N]
        * spin unpolarized
            LDA : [1,1,1,N]
            GGA : [4,4,4,N]
            MGGA: [5,5,5,N]
    '''
    rho = np.asarray(rho, order='C')
    if xctype == 'GGA':
        order = 1
        nvar = 4
        frg = fxc[1].T
        fgg = fxc[2].T
        frrr, frrg, frgg, fggg = [x.T for x in kxc[:4]]
    elif xctype == 'MGGA':
        order = 2
        nvar = 5
        frg = fxc[1].T
        fgg = fxc[2].T
        fgt = fxc[9].T
        frrr, frrg, frgg, fggg, frrt, frgt, frtt, fggt, fgtt, fttt = \
                [kxc[i].T for i in [0, 1, 2, 3, 5, 7, 10, 12, 15, 19]]
    else:  # LDA
        order = 0
        nvar = 1
        frrr = kxc[0].T

    ngrids = rho.shape[-1]
    if spin == 1:
        if order == 0:
            vp = _stack_frrr(frrr).reshape(2,nvar, 2,nvar, 2,nvar, ngrids).transpose(1,3,5,0,2,4,6)
        else:
            vp = np.empty((2,nvar, 2,nvar, 2,nvar, ngrids)).transpose(1,3,5,0,2,4,6)
            vp[0,0,0] = _stack_frrr(frrr)
            i3 = np.arange(3)
            #:qggg = _stack_fggg(fggg)
            #:qggg = np.einsum('abcdefg,axg->xbcdefg', qggg, rho[:,1:4])
            #:qggg = np.einsum('xbcdefg,cyg->xybdefg', qggg, rho[:,1:4])
            #:qggg = np.einsum('xybdefg,ezg->xyzbdfg', qggg, rho[:,1:4])
            qggg = _stack_fggg(fggg, rho=rho).transpose(1,3,5,0,2,4,6)
            qgg = _stack_fgg(fgg)
            qgg = np.einsum('abcdg,axg->xbcdg', qgg, rho[:,1:4])
            for i in range(3):
                qggg[:,i,i] += qgg
                qggg[i,:,i] += qgg.transpose(0,2,1,3,4)
                qggg[i,i,:] += qgg.transpose(0,2,3,1,4)
            vp[1:4,1:4,1:4] = qggg

            frgg = frgg.reshape(2,6,ngrids)
            #:qrgg = _stack_fgg(frgg, axis=1)
            #:qrgg = np.einsum('rabcdg,axg->xrbcdg', qrgg, rho[:,1:4])
            #:qrgg = np.einsum('xrbcdg,cyg->xyrbdg', qrgg, rho[:,1:4])
            qrgg = _stack_fgg(frgg, axis=1, rho=rho).transpose(2,4,0,1,3,5)
            qrg = _stack_fg(frg.reshape(2,3,ngrids), axis=1)
            qrgg[i3,i3] += qrg
            vp[0,1:4,1:4] = qrgg
            vp[1:4,0,1:4] = qrgg.transpose(0,1,3,2,4,5)
            vp[1:4,1:4,0] = qrgg.transpose(0,1,3,4,2,5)

            frrg = frrg.reshape(3,3,ngrids)
            qrrg = _stack_frr(frrg, axis=0)
            #:qrrg = _stack_fg(qrrg, axis=2)
            #:qrrg = np.einsum('rsabg,axg->rsxbg', qrrg, rho[:,1:4])
            qrrg = _stack_fg(qrrg, axis=2, rho=rho).transpose(3,0,1,2,4)
            vp[0,0,1:4] = qrrg
            vp[0,1:4,0] = qrrg.transpose(0,1,3,2,4)
            vp[1:4,0,0] = qrrg.transpose(0,3,1,2,4)

        if order > 1:
            fggt = fggt.reshape(6,2,ngrids)
            #:qggt = _stack_fgg(fggt, axis=0)
            #:qggt = np.einsum('abcdrg,axg->xbcdrg', qggt, rho[:,1:4])
            #:qggt = np.einsum('xbcdrg,cyg->xybdrg', qggt, rho[:,1:4])
            qggt = _stack_fgg(fggt, axis=0, rho=rho).transpose(1,3,0,2,4,5)
            qgt = _stack_fg(fgt.reshape(3,2,ngrids), axis=0)
            i3 = np.arange(3)
            qggt[i3,i3] += qgt
            vp[1:4,1:4,4] = qggt
            vp[1:4,4,1:4] = qggt.transpose(0,1,2,4,3,5)
            vp[4,1:4,1:4] = qggt.transpose(0,1,4,2,3,5)

            qgtt = _stack_frr(fgtt.reshape(3,3,ngrids), axis=1)
            #:qgtt = _stack_fg(qgtt, axis=0)
            #:qgtt = np.einsum('abrsg,axg->xbrsg', qgtt, rho[:,1:4])
            qgtt = _stack_fg(qgtt, axis=0, rho=rho).transpose(1,0,2,3,4)
            vp[1:4,4,4] = qgtt
            vp[4,1:4,4] = qgtt.transpose(0,2,1,3,4)
            vp[4,4,1:4] = qgtt.transpose(0,2,3,1,4)

            frgt = frgt.reshape(2,3,2,ngrids)
            #:qrgt = _stack_fg(frgt, axis=1)
            #:qrgt = np.einsum('rabsg,axg->xrbsg', qrgt, rho[:,1:4])
            qrgt = _stack_fg(frgt, axis=1, rho=rho).transpose(2,0,1,3,4)
            vp[0,1:4,4] = qrgt
            vp[0,4,1:4] = qrgt.transpose(0,1,3,2,4)
            vp[1:4,0,4] = qrgt.transpose(0,2,1,3,4)
            vp[4,0,1:4] = qrgt.transpose(0,3,1,2,4)
            vp[1:4,4,0] = qrgt.transpose(0,2,3,1,4)
            vp[4,1:4,0] = qrgt.transpose(0,3,2,1,4)

            qrrt = _stack_frr(frrt.reshape(3,2,ngrids), axis=0)
            vp[0,0,4] = qrrt
            vp[0,4,0] = qrrt.transpose(0,2,1,3)
            vp[4,0,0] = qrrt.transpose(2,0,1,3)

            qrtt = _stack_frr(frtt.reshape(2,3,ngrids), axis=1)
            vp[0,4,4] = qrtt
            vp[4,0,4] = qrtt.transpose(1,0,2,3)
            vp[4,4,0] = qrtt.transpose(1,2,0,3)

            vp[4,4,4] = _stack_frrr(fttt, axis=0)

        vp = vp.transpose(3,0,4,1,5,2,6)

    else:
        if order == 0:
            vp = frrr.reshape(nvar, nvar, nvar, ngrids)
        else:
            vp = np.empty((nvar, nvar, nvar, ngrids))
            vp[0,0,0] = frrr
            i3 = np.arange(3)
            qggg = 8 * fggg * rho[1:4] * rho[1:4,None] * rho[1:4,None,None]
            qgg = 4 * fgg * rho[1:4]
            for i in range(3):
                qggg[i,i,:] += qgg
                qggg[i,:,i] += qgg
                qggg[:,i,i] += qgg
            vp[1:4,1:4,1:4] = qggg

            qrgg = 4 * frgg * rho[1:4] * rho[1:4,None]
            qrgg[i3,i3] += frg * 2
            vp[0,1:4,1:4] = qrgg
            vp[1:4,0,1:4] = qrgg
            vp[1:4,1:4,0] = qrgg

            qrrg = 2 * frrg * rho[1:4]
            vp[0,0,1:4] = qrrg
            vp[0,1:4,0] = qrrg
            vp[1:4,0,0] = qrrg

        if order > 1:
            qggt = 4 * fggt * rho[1:4] * rho[1:4,None]
            qggt[i3,i3] += fgt * 2
            vp[1:4,1:4,4] = qggt
            vp[1:4,4,1:4] = qggt
            vp[4,1:4,1:4] = qggt

            qgtt = 2 * fgtt * rho[1:4]
            vp[1:4,4,4] = qgtt
            vp[4,1:4,4] = qgtt
            vp[4,4,1:4] = qgtt

            qrgt = 2 * frgt * rho[1:4]
            vp[0,1:4,4] = qrgt
            vp[0,4,1:4] = qrgt
            vp[1:4,0,4] = qrgt
            vp[4,0,1:4] = qrgt
            vp[1:4,4,0] = qrgt
            vp[4,1:4,0] = qrgt

            vp[0,0,4] = frrt
            vp[0,4,0] = frrt
            vp[4,0,0] = frrt
            vp[0,4,4] = frtt
            vp[4,0,4] = frtt
            vp[4,4,0] = frtt
            vp[4,4,4] = fttt
    return vp

def transform_lxc(rho, fxc, kxc, lxc, xctype, spin=0):
    r'''
    Transform libxc vxc functional output to the derivative tensor of parameters
    in rho: [density, nabla_x, nabla_y, nabla_z, tau]. The output tensor has the
    shape:
        * spin polarized
            LDA : [2,1,2,1,2,1,2,1,N]
            GGA : [2,4,2,4,2,4,2,4,N]
            MGGA: [2,5,2,5,2,5,2,5,N]
        * spin unpolarized
            LDA : [1,1,1,1,N]
            GGA : [4,4,4,4,N]
            MGGA: [5,5,5,5,N]
    '''
    raise NotImplementedError

def _stack_fg(fg, axis=0, rho=None, out=None):
    '''fg [uu, ud, dd] -> [[uu*2, ud], [du, dd*2]]'''
    if rho is None:
        qg = _stack_frr(fg, axis)
        if axis == 0:
            qg[0,0] *= 2
            qg[1,1] *= 2
        elif axis == 1:
            qg[:,0,0] *= 2
            qg[:,1,1] *= 2
        elif axis == 2:
            qg[:,:,0,0] *= 2
            qg[:,:,1,1] *= 2
        else:
            raise NotImplementedError
        return qg
    else:
        rho = np.asarray(rho, order='C')
        fg = np.asarray(fg, order='C')
        if fg.shape[axis] != 3:
            fg = fg.reshape(fg.shape[:axis] + (3, -1) + fg.shape[axis+1:])
        nvar, ngrids = rho.shape[1:]
        ncounts = int(np.prod(fg.shape[:axis]))
        mcounts = int(np.prod(fg.shape[axis+1:-1]))
        qg = np.empty(fg.shape[:axis] + (2, 3) + fg.shape[axis+1:])
        rho = libdft.VXCfg_to_direct_deriv(
            qg.ctypes.data_as(ctypes.c_void_p),
            fg.ctypes.data_as(ctypes.c_void_p),
            rho.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(ncounts), ctypes.c_int(nvar), ctypes.c_int(mcounts),
            ctypes.c_int(ngrids))
        return qg


def _stack_frr(frr, axis=0):
    '''frr [u_u, u_d, d_d] -> [[u_u, u_d], [d_u, d_d]]'''
    if frr.shape[axis] != 3:
        frr = frr.reshape(frr.shape[:axis] + (3, -1) + frr.shape[axis+1:])
    slices = [slice(None)] * frr.ndim
    slices[axis] = [[0,1],[1,2]]
    return frr[tuple(slices)]

def _stack_fgg(fgg, axis=0, rho=None):
    '''
    fgg [uu_uu, uu_ud, uu_dd, ud_ud, ud_dd, dd_dd] ->
    [[uu_uu, ud_ud, ud_dd],
     [ud_uu, ud_ud, ud_dd],
     [dd_uu, dd_ud, dd_dd]] -> tensor with shape [2,2, 2,2, ...]
    '''
    if fgg.shape[axis] != 6:
        fgg = fgg.reshape(fgg.shape[:axis] + (6, -1) + fgg.shape[axis+1:])
    slices = [slice(None)] * fgg.ndim
    slices[axis] = [[0, 1, 2], [1, 3, 4], [2, 4, 5]]
    fgg = fgg[tuple(slices)]
    return _stack_fg(_stack_fg(fgg, axis=axis+1, rho=rho), axis=axis, rho=rho)

def _stack_frrr(frrr, axis=0):
    '''
    frrr [u_u_u, u_u_d, u_d_d, d_d_d]
    -> tensor with shape [2, 2, 2, ...]
    '''
    if frrr.shape[axis] != 4:
        frrr = frrr.reshape(frrr.shape[:axis] + (4, -1) + frrr.shape[axis+1:])
    slices = [slice(None)] * frrr.ndim
    slices[axis] = [[[0, 1], [1, 2]],
                    [[1, 2], [2, 3]]]
    return frrr[tuple(slices)]

def _stack_fggg(fggg, axis=0, rho=None):
    '''
    fggg [uu_uu_uu, uu_uu_ud, uu_uu_dd, uu_ud_ud, uu_ud_dd, uu_dd_dd, ud_ud_ud, ud_ud_dd, ud_dd_dd, dd_dd_dd]
    -> tensor with shape [2,2, 2,2, 2,2, ...]
    '''
    if fggg.shape[axis] != 10:
        fggg = fggg.reshape(fggg.shape[:axis] + (10, 2) + fggg.shape[axis+1:])
    slices = [slice(None)] * fggg.ndim
    slices[axis] = [[[0, 1, 2], [1, 3, 4], [2, 4, 5]],
                    [[1, 3, 4], [3, 6, 7], [4, 7, 8]],
                    [[2, 4, 5], [4, 7, 8], [5, 8, 9]]]
    fggg = fggg[tuple(slices)]
    fggg = _stack_fg(fggg, axis=axis+2, rho=rho)
    fggg = _stack_fg(fggg, axis=axis+1, rho=rho)
    return _stack_fg(fggg, axis=axis, rho=rho)

@lru_cache(100)
def _product_uniq_indices(nvars, order):
    '''
    Indexing the symmetry unique elements in cartesian product
    '''
    # n = 0
    # for i range(nvars):
    #    for j in range(i, nvars):
    #        for k in range(k, nvars):
    #            ...
    #            prod[(i,j,k,...)] = n
    #            n += 1
    prod = np.ones([nvars] * order, dtype=int)
    uniq_idx = list(itertools.combinations_with_replacement(range(nvars), order))
    prod[tuple(np.array(uniq_idx).T)] = np.arange(len(uniq_idx))

    idx = np.where(prod)
    sidx = np.sort(idx, axis=0)
    prod[idx] = prod[tuple(sidx)]
    return prod

def _pair_combinations(lst):
    n = len(lst)
    if n <= 2:
        return [[tuple(lst)]]
    a = lst[0]
    results = []
    for i in range(1, n):
        pair = (a, lst[i])
        rests = _pair_combinations(lst[1:i]+lst[i+1:])
        for rest in rests:
            rest.append(pair)
        results.extend(rests)
    return results

def _unfold_gga(rho, xc_val, spin, order, nvar, xlen, reserve=0):
    assert nvar >= 4
    ngrids = rho.shape[-1]
    n_transform = order - reserve

    if spin == 0:
        nvar2 = nvar
        drv = libdft.VXCunfold_sigma_spin0
    else:
        nvar2 = nvar * 2
        drv = libdft.VXCunfold_sigma_spin1

    # xc_val[idx] expands unique xc elements to n-order tensors
    idx = _product_uniq_indices(xlen, order)
    xc_tensor = np.empty((xlen**reserve * nvar2**n_transform, ngrids))
    xc_tensor[:idx.size] = xc_val[idx.ravel()]
    buf = np.empty_like(xc_tensor)
    for i in range(n_transform):
        xc_tensor, buf = buf, xc_tensor
        ncounts = xlen**(order-1-i) * nvar2**i
        drv(xc_tensor.ctypes, buf.ctypes, rho.ctypes,
            ctypes.c_int(ncounts), ctypes.c_int(nvar), ctypes.c_int(ngrids))
    if spin == 0:
        xc_tensor = xc_tensor.reshape([xlen]*reserve + [nvar]*n_transform + [ngrids])
    else:
        xc_tensor = xc_tensor.reshape([xlen]*reserve + [2,nvar]*n_transform + [ngrids])
    return xc_tensor

def _diagonal_indices(idx, order):
    assert order > 0
    n = len(idx)
    diag_idx = [np.asarray(idx)]
    for i in range(1, order):
        last_dim = diag_idx[-1]
        diag_idx = [np.repeat(x, n) for x in diag_idx]
        diag_idx.append(np.tile(last_dim, n))
    # repeat(diag_idx, 2)
    return tuple(x for x in diag_idx for i in range(2))

_XC_NVAR = {
    ('HF', 0): (1, 1),
    ('HF', 1): (1, 2),
    ('LDA', 0): (1, 1),
    ('LDA', 1): (1, 2),
    ('GGA', 0): (4, 2),
    ('GGA', 1): (4, 5),
    ('MGGA', 0): (5, 3),
    ('MGGA', 1): (5, 7),
}

def transform_xc(rho, xc_val, xctype, spin, order):
    '''General transformation to construct XC derivative tensor'''
    xc_val = np.asarray(xc_val, order='C')
    if order == 0:
        return xc_val[0]

    nvar, xlen = _XC_NVAR[xctype, spin]
    ngrids = xc_val.shape[-1]
    offsets = [0] + [count_combinations(xlen, o) for o in range(order+1)]
    p0, p1 = offsets[order:order+2]
    if xctype == 'LDA' or xctype == 'HF':
        xc_out = xc_val[p0:p1]
        if spin == 0:
            return xc_out.reshape([1]*order + [ngrids])
        else:
            idx = _product_uniq_indices(xlen, order)
            return xc_out[idx].reshape([2,1]*order + [ngrids])

    rho = np.asarray(rho, order='C')
    assert rho.size == (spin+1) * nvar * ngrids
    xc_tensor = _unfold_gga(rho, xc_val[p0:p1], spin, order, nvar, xlen)
    nabla_idx = np.arange(1, 4)
    if spin == 0:
        for n_pairs in range(1, order//2+1):
            p0, p1 = offsets[order-n_pairs:order-n_pairs+2]
            xc_sub = _unfold_gga(rho, xc_val[p0:p1], spin, order-n_pairs,
                                 nvar, xlen, n_pairs)
            # Just the sigma components
            xc_sub = xc_sub[(1,)*n_pairs] * 2**n_pairs

            # low_sigmas indicates the index for the extra sigma terms
            low_sigmas = itertools.combinations(range(order), n_pairs*2)
            pair_combs = [list(itertools.chain(*p[::-1]))
                          for p in _pair_combinations(list(range(n_pairs*2)))]
            diag_idx = _diagonal_indices(nabla_idx, n_pairs)

            for dim_lst in low_sigmas:
                rest_dims = [i for i in range(xc_tensor.ndim) if i not in dim_lst]
                for pair_comb in pair_combs:
                    xc_tensor_1 = xc_tensor.transpose(
                        [dim_lst[i] for i in pair_comb] + rest_dims)
                    xc_tensor_1[diag_idx] += xc_sub
    else:
        i3to2x2 = _product_uniq_indices(2, 2)
        for n_pairs in range(1, order//2+1):
            p0, p1 = offsets[order-n_pairs:order-n_pairs+2]
            xc_sub = _unfold_gga(rho, xc_val[p0:p1], spin, order-n_pairs,
                                 nvar, xlen, n_pairs)
            # Just the sigma components
            xc_sub = xc_sub[(slice(2,5),) * n_pairs]
            for i in range(n_pairs):
                xc_sub[(slice(None),)*i+(0,)] *= 2
                xc_sub[(slice(None),)*i+(2,)] *= 2
            sigma_idx = (i3to2x2[(slice(None),)*2 + (np.newaxis,)*(i*2)]
                         for i in reversed(range(n_pairs)))
            xc_sub = xc_sub[tuple(sigma_idx)]

            low_sigmas = itertools.combinations(range(order), n_pairs*2)
            pair_combs = [list(itertools.chain(*p[::-1]))
                          for p in _pair_combinations(list(range(n_pairs*2)))]
            diag_idx = _diagonal_indices(nabla_idx, n_pairs)

            for dim_lst in low_sigmas:
                dim_lst2 = np.array(dim_lst)*2 + np.array([1, 0])[:,None]
                rest_dims = [i for i in range(xc_tensor.ndim) if i not in dim_lst2]
                for pair_comb in pair_combs:
                    leading_dims = list(dim_lst2[:,pair_comb].ravel())
                    xc_tensor_1 = xc_tensor.transpose(leading_dims + rest_dims)
                    xc_tensor_1[diag_idx] += xc_sub
    return xc_tensor

def count_combinations(nvar, order):
    '''sum(len(combinations_with_replacement(range(nvar), o) for o in range(order))'''
    return lib.comb(nvar+order, order)

def ud2ts(v_ud):
    '''XC derivatives spin-up/spin-down representations to
    total-density/spin-density representations'''
    v_ud = np.asarray(v_ud, order='C')
    v_ts = np.empty_like(v_ud)
    nvar, ngrids = v_ts.shape[-2:]
    nvar2 = nvar * 2
    order = v_ud.ndim // 2
    drv = libdft.VXCud2ts
    v_ts, v_ud = v_ud, v_ts
    for i in range(order):
        v_ts, v_ud = v_ud, v_ts
        n = nvar2**i
        nvg = nvar * nvar2**(order-1-i) * ngrids
        #:c = np.array([[.5,  .5],    # vrho = (va + vb) / 2
        #:              [.5, -.5]])   # vs   = (va - vb) / 2
        #:v_ts = np.einsum('ra,...ax...g->...rx...g', c, v_ud)
        drv(v_ts.ctypes, v_ud.ctypes, ctypes.c_int(n), ctypes.c_int(nvg))
    return v_ts

def ts2ud(v_ts):
    '''XC derivatives total-density/spin-density representations to
    spin-up/spin-down representations'''
    v_ts = np.asarray(v_ts, order='C')
    v_ud = np.empty_like(v_ts)
    nvar, ngrids = v_ts.shape[-2:]
    nvar2 = nvar * 2
    order = v_ud.ndim // 2
    drv = libdft.VXCts2ud
    v_ts, v_ud = v_ud, v_ts
    for i in range(order):
        v_ts, v_ud = v_ud, v_ts
        n = nvar2**i
        nvg = nvar * nvar2**(order-1-i) * ngrids
        #:c = np.array([[1,  1],    # va = vrho + vs
        #:              [1, -1]])   # vb = vrho - vs
        #:v_ud = np.einsum('ra,...ax...g->...rx...g', c, v_ts)
        drv(v_ud.ctypes, v_ts.ctypes, ctypes.c_int(n), ctypes.c_int(nvg))
    return v_ud



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#
# Ref:
# J. Chem. Phys. 117, 7433
#


from functools import reduce
import numpy
from pyscf import gto
from pyscf import lib
from pyscf.lib import logger
from pyscf.grad import rhf as rhf_grad
from pyscf.scf import cphf
from pyscf import __config__


def grad_elec(td_grad, x_y, singlet=True, atmlst=None,
              max_memory=2000, verbose=logger.INFO):
    '''
    Electronic part of TDA, TDHF nuclear gradients

    Args:
        td_grad : grad.tdrhf.Gradients or grad.tdrks.Gradients object.

        x_y : a two-element list of numpy arrays
            TDDFT X and Y amplitudes. If Y is set to 0, this function computes
            TDA energy gradients.
    '''
    log = logger.new_logger(td_grad, verbose)
    time0 = logger.process_clock(), logger.perf_counter()

    assert td_grad.base.frozen is None

    mol = td_grad.mol
    mf = td_grad.base._scf
    mo_coeff = mf.mo_coeff
    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    nao, nmo = mo_coeff.shape
    nocc = (mo_occ>0).sum()
    nvir = nmo - nocc
    x, y = x_y
    xpy = (x+y).reshape(nocc,nvir).T
    xmy = (x-y).reshape(nocc,nvir).T
    orbv = mo_coeff[:,nocc:]
    orbo = mo_coeff[:,:nocc]

    dvv = numpy.einsum('ai,bi->ab', xpy, xpy) + numpy.einsum('ai,bi->ab', xmy, xmy)
    doo =-numpy.einsum('ai,aj->ij', xpy, xpy) - numpy.einsum('ai,aj->ij', xmy, xmy)
    dmxpy = reduce(numpy.dot, (orbv, xpy, orbo.T))
    dmxmy = reduce(numpy.dot, (orbv, xmy, orbo.T))
    dmzoo = reduce(numpy.dot, (orbo, doo, orbo.T))
    dmzoo+= reduce(numpy.dot, (orbv, dvv, orbv.T))

    vj, vk = mf.get_jk(mol, (dmzoo, dmxpy+dmxpy.T, dmxmy-dmxmy.T), hermi=0)
    veff0doo = vj[0] * 2 - vk[0]
    wvo = reduce(numpy.dot, (orbv.T, veff0doo, orbo)) * 2
    if singlet:
        veff = vj[1] * 2 - vk[1]
    else:
        veff = -vk[1]
    veff0mop = reduce(numpy.dot, (mo_coeff.T, veff, mo_coeff))
    wvo -= numpy.einsum('ki,ai->ak', veff0mop[:nocc,:nocc], xpy) * 2
    wvo += numpy.einsum('ac,ai->ci', veff0mop[nocc:,nocc:], xpy) * 2
    veff = -vk[2]
    veff0mom = reduce(numpy.dot, (mo_coeff.T, veff, mo_coeff))
    wvo -= numpy.einsum('ki,ai->ak', veff0mom[:nocc,:nocc], xmy) * 2
    wvo += numpy.einsum('ac,ai->ci', veff0mom[nocc:,nocc:], xmy) * 2

    # set singlet=None, generate function for CPHF type response kernel
    vresp = mf.gen_response(singlet=None, hermi=1,
                            with_nlc=not td_grad.base.exclude_nlc)
    def fvind(x):  # For singlet, closed shell ground state
        dm = reduce(numpy.dot, (orbv, x.reshape(nvir,nocc)*2, orbo.T))
        v1ao = vresp(dm+dm.T)
        return reduce(numpy.dot, (orbv.T, v1ao, orbo)).ravel()
    z1 = cphf.solve(fvind, mo_energy, mo_occ, wvo,
                    max_cycle=td_grad.cphf_max_cycle,
                    tol=td_grad.cphf_conv_tol)[0]
    z1 = z1.reshape(nvir,nocc)
    time1 = log.timer('Z-vector using CPHF solver', *time0)

    z1ao = reduce(numpy.dot, (orbv, z1, orbo.T))
    veff = vresp(z1ao+z1ao.T)

    im0 = numpy.zeros((nmo,nmo))
    im0[:nocc,:nocc] = reduce(numpy.dot, (orbo.T, veff0doo+veff, orbo))
    im0[:nocc,:nocc]+= numpy.einsum('ak,ai->ki', veff0mop[nocc:,:nocc], xpy)
    im0[:nocc,:nocc]+= numpy.einsum('ak,ai->ki', veff0mom[nocc:,:nocc], xmy)
    im0[nocc:,nocc:] = numpy.einsum('ci,ai->ac', veff0mop[nocc:,:nocc], xpy)
    im0[nocc:,nocc:]+= numpy.einsum('ci,ai->ac', veff0mom[nocc:,:nocc], xmy)
    im0[nocc:,:nocc] = numpy.einsum('ki,ai->ak', veff0mop[:nocc,:nocc], xpy)*2
    im0[nocc:,:nocc]+= numpy.einsum('ki,ai->ak', veff0mom[:nocc,:nocc], xmy)*2

    zeta = lib.direct_sum('i+j->ij', mo_energy, mo_energy) * .5
    zeta[nocc:,:nocc] = mo_energy[:nocc]
    zeta[:nocc,nocc:] = mo_energy[nocc:]
    dm1 = numpy.zeros((nmo,nmo))
    dm1[:nocc,:nocc] = doo
    dm1[nocc:,nocc:] = dvv
    dm1[nocc:,:nocc] = z1
    dm1[:nocc,:nocc] += numpy.eye(nocc)*2 # for ground state
    im0 = reduce(numpy.dot, (mo_coeff, im0+zeta*dm1, mo_coeff.T))

    # Initialize hcore_deriv with the underlying SCF object because some
    # extensions (e.g. QM/MM, solvent) modifies the SCF object only.
    mf_grad = td_grad.base._scf.nuc_grad_method()
    hcore_deriv = mf_grad.hcore_generator(mol)
    s1 = mf_grad.get_ovlp(mol)

    dmz1doo = z1ao + dmzoo
    oo0 = reduce(numpy.dot, (orbo, orbo.T))
    vj, vk = td_grad.get_jk(mol, (oo0, dmz1doo+dmz1doo.T, dmxpy+dmxpy.T,
                                  dmxmy-dmxmy.T))
    vj = vj.reshape(-1,3,nao,nao)
    vk = vk.reshape(-1,3,nao,nao)
    vhf1 = -vk
    if singlet:
        vhf1 += vj * 2
    else:
        vhf1[:2] += vj[:2]*2
    time1 = log.timer('2e AO integral derivatives', *time1)

    if atmlst is None:
        atmlst = range(mol.natm)
    offsetdic = mol.offset_nr_by_atom()
    de = numpy.zeros((len(atmlst),3))
    for k, ia in enumerate(atmlst):
        shl0, shl1, p0, p1 = offsetdic[ia]

        # Ground state gradients
        h1ao = hcore_deriv(ia)
        h1ao[:,p0:p1]   += vhf1[0,:,p0:p1]
        h1ao[:,:,p0:p1] += vhf1[0,:,p0:p1].transpose(0,2,1)
        # oo0*2 for doubly occupied orbitals
        de[k] = numpy.einsum('xpq,pq->x', h1ao, oo0) * 2
        de[k] += numpy.einsum('xpq,pq->x', h1ao, dmz1doo)

        de[k] -= numpy.einsum('xpq,pq->x', s1[:,p0:p1], im0[p0:p1])
        de[k] -= numpy.einsum('xqp,pq->x', s1[:,p0:p1], im0[:,p0:p1])

        de[k] += numpy.einsum('xij,ij->x', vhf1[1,:,p0:p1], oo0[p0:p1])
        de[k] += numpy.einsum('xij,ij->x', vhf1[2,:,p0:p1], dmxpy[p0:p1,:]) * 2
        de[k] += numpy.einsum('xij,ij->x', vhf1[3,:,p0:p1], dmxmy[p0:p1,:]) * 2
        de[k] += numpy.einsum('xji,ij->x', vhf1[2,:,p0:p1], dmxpy[:,p0:p1]) * 2
        de[k] -= numpy.einsum('xji,ij->x', vhf1[3,:,p0:p1], dmxmy[:,p0:p1]) * 2

        de[k] += td_grad.extra_force(ia, locals())

    log.timer('TDHF nuclear gradients', *time0)
    return de


def as_scanner(td_grad, state=1):
    '''Generating a nuclear gradients scanner/solver (for geometry optimizer).

    The returned solver is a function. This function requires one argument
    "mol" as input and returns energy and first order nuclear derivatives.

    The solver will automatically use the results of last calculation as the
    initial guess of the new calculation.  All parameters assigned in the
    nuc-grad object and SCF object (DIIS, conv_tol, max_memory etc) are
    automatically applied in the solver.

    Note scanner has side effects.  It may change many underlying objects
    (_scf, with_df, with_x2c, ...) during calculation.

    Examples::

    >>> from pyscf import gto, scf, tdscf, grad
    >>> mol = gto.M(atom='H 0 0 0; F 0 0 1')
    >>> td_grad_scanner = scf.RHF(mol).apply(tdscf.TDA).nuc_grad_method().as_scanner()
    >>> e_tot, grad = td_grad_scanner(gto.M(atom='H 0 0 0; F 0 0 1.1'))
    >>> e_tot, grad = td_grad_scanner(gto.M(atom='H 0 0 0; F 0 0 1.5'))
    '''
    from pyscf import gto
    if isinstance(td_grad, lib.GradScanner):
        return td_grad

    if state == 0:
        return td_grad.base._scf.nuc_grad_method().as_scanner()

    logger.info(td_grad, 'Create scanner for %s', td_grad.__class__)
    name = td_grad.__class__.__name__ + TDSCF_GradScanner.__name_mixin__
    return lib.set_class(TDSCF_GradScanner(td_grad, state),
                         (TDSCF_GradScanner, td_grad.__class__), name)

class TDSCF_GradScanner(lib.GradScanner):
    _keys = {'e_tot'}

    def __init__(self, g, state):
        lib.GradScanner.__init__(self, g)
        if state is not None:
            self.state = state

    def __call__(self, mol_or_geom, state=None, **kwargs):
        if isinstance(mol_or_geom, gto.MoleBase):
            assert mol_or_geom.__class__ == gto.Mole
            mol = mol_or_geom
        else:
            mol = self.mol.set_geom_(mol_or_geom, inplace=False)
        self.reset(mol)

        if state is None:
            state = self.state
        else:
            self.state = state

        td_scanner = self.base
        td_scanner(mol)
# TODO: Check root flip.  Maybe avoid the initial guess in TDHF otherwise
# large error may be found in the excited states amplitudes
        de = self.kernel(state=state, **kwargs)
        e_tot = self.e_tot[state-1]
        return e_tot, de

    @property
    def converged(self):
        td_scanner = self.base
        return all((td_scanner._scf.converged,
                    td_scanner.converged[self.state]))


class Gradients(rhf_grad.GradientsBase):

    cphf_max_cycle = getattr(__config__, 'grad_tdrhf_Gradients_cphf_max_cycle', 20)
    cphf_conv_tol = getattr(__config__, 'grad_tdrhf_Gradients_cphf_conv_tol', 1e-8)

    _keys = {
        'cphf_max_cycle', 'cphf_conv_tol',
        'mol', 'base', 'chkfile', 'state', 'atmlst', 'de',
    }

    def __init__(self, td):
        self.verbose = td.verbose
        self.stdout = td.stdout
        self.mol = td.mol
        self.base = td
        self.chkfile = td.chkfile
        self.max_memory = td.max_memory
        self.state = 1  # of which the gradients to be computed.
        self.atmlst = None
        self.de = None

        if getattr(td._scf, 'with_df', None):
            raise NotImplementedError('Nuclear Gradients for DF-TDDFT')

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('\n')
        log.info('******** LR %s gradients for %s ********',
                 self.base.__class__, self.base._scf.__class__)
        log.info('cphf_conv_tol = %g', self.cphf_conv_tol)
        log.info('cphf_max_cycle = %d', self.cphf_max_cycle)
        log.info('chkfile = %s', self.chkfile)
        log.info('State ID = %d', self.state)
        log.info('max_memory %d MB (current use %d MB)',
                 self.max_memory, lib.current_memory()[0])
        log.info('\n')
        return self

    @lib.with_doc(grad_elec.__doc__)
    def grad_elec(self, xy, singlet, atmlst=None):
        return grad_elec(self, xy, singlet, atmlst, self.max_memory, self.verbose)

    def kernel(self, xy=None, state=None, singlet=None, atmlst=None):
        '''
        Args:
            state : int
                Excited state ID.  state = 1 means the first excited state.
        '''
        if xy is None:
            if state is None:
                state = self.state
            else:
                self.state = state

            if state == 0:
                logger.warn(self, 'state=0 found in the input. '
                            'Gradients of ground state is computed.')
                return self.base._scf.nuc_grad_method().kernel(atmlst=atmlst)

            if self.base.xy is None:
                self.base.run()
            xy = self.base.xy[state-1]

        if singlet is None: singlet = self.base.singlet
        if atmlst is None:
            atmlst = self.atmlst
        else:
            self.atmlst = atmlst

        if self.verbose >= logger.WARN:
            self.check_sanity()
        if self.verbose >= logger.INFO:
            self.dump_flags()

        de = self.grad_elec(xy, singlet, atmlst)
        self.de = de = de + self.grad_nuc(atmlst=atmlst)
        if self.mol.symmetry:
            self.de = self.symmetrize(self.de, atmlst)
        self._finalize()
        return self.de

    # Calling the underlying SCF nuclear gradients because it may be modified
    # by external modules (e.g. QM/MM, solvent)
    def grad_nuc(self, mol=None, atmlst=None):
        mf_grad = self.base._scf.nuc_grad_method()
        return mf_grad.grad_nuc(mol, atmlst)

    def _finalize(self):
        if self.verbose >= logger.NOTE:
            logger.note(self, '--------- %s gradients for state %d ----------',
                        self.base.__class__.__name__, self.state)
            self._write(self.mol, self.de, self.atmlst)
            logger.note(self, '----------------------------------------------')

    as_scanner = as_scanner

    to_gpu = lib.to_gpu

Grad = Gradients

from pyscf import tdscf
tdscf.rhf.TDA.Gradients = tdscf.rhf.TDHF.Gradients = lib.class_as_method(Gradients)
